# Copyright (c) 2023 Canonical Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import copy

import netaddr

from neutron_lib.api.definitions import external_net as enet_apidef
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import l3_ext_gw_mode
from neutron_lib.api.definitions import l3_ext_gw_multihoming
from neutron_lib import constants
from neutron_lib.db import api as db_api
from neutron_lib import exceptions
from neutron_lib.utils import net as net_utils
from oslo_utils import uuidutils

from neutron.db import l3_extra_gws_db
from neutron.db.models import l3 as l3_models
from neutron.ipam import exceptions as ipam_exceptions
from neutron.objects import ipam as ipam_obj
from neutron.objects import network as net_obj
from neutron.objects import ports as port_obj
from neutron.objects import router as l3_obj
from neutron.objects import subnet as subnet_obj
from neutron.tests.unit.extensions import test_l3
from neutron.tests.unit import testlib_api

_uuid = uuidutils.generate_uuid


class TestDbIntPlugin(test_l3.TestL3NatIntPlugin,
                      l3_extra_gws_db.ExtraGatewaysMixinDbMixin):

    supported_extension_aliases = [enet_apidef.ALIAS, l3_apidef.ALIAS,
                                   l3_ext_gw_mode.ALIAS,
                                   l3_ext_gw_multihoming.ALIAS]


class TestExtraGatewaysDb(testlib_api.SqlTestCase):

    def setUp(self):
        super().setUp()
        plugin = __name__ + '.' + TestDbIntPlugin.__name__
        self.setup_coreplugin(plugin)
        self.target_object = TestDbIntPlugin()
        # Patch the context
        ctx_patcher = mock.patch('neutron_lib.context', autospec=True)
        mock_context = ctx_patcher.start()
        self.context = mock_context.get_admin_context()
        self.context.elevated.return_value = self.context
        self.context.session = db_api.get_writer_session()

        # Create a simple setup with one external network and a subnet on it.
        self.ext_net_a_id = _uuid()
        self.ext_sub_a_id = _uuid()
        self.tenant_id = _uuid()

        self.network_a = net_obj.Network(
            self.context,
            id=self.ext_net_a_id,
            project_id=self.tenant_id,
            admin_state_up=True,
            status=constants.NET_STATUS_ACTIVE)
        self.network_a.create()
        self.net_ext_a = net_obj.ExternalNetwork(
            self.context, network_id=self.ext_net_a_id)
        self.net_ext_a.create()
        self.ext_sub_a = subnet_obj.Subnet(self.context,
                                           id=self.ext_sub_a_id,
            project_id=self.tenant_id,
            ip_version=constants.IP_VERSION_4,
            cidr=net_utils.AuthenticIPNetwork('192.0.2.0/25'),
            gateway_ip=netaddr.IPAddress('192.0.2.1'),
            network_id=self.ext_net_a_id)
        self.ext_sub_a.create()

        self.ext_net_b_id = _uuid()
        self.ext_sub_b_id = _uuid()
        self.network_b = net_obj.Network(
            self.context,
            id=self.ext_net_b_id,
            project_id=self.tenant_id,
            admin_state_up=True,
            status=constants.NET_STATUS_ACTIVE)
        self.network_b.create()
        self.net_ext_b = net_obj.ExternalNetwork(
            self.context, network_id=self.ext_net_b_id)
        self.net_ext_b.create()

        self.ext_sub_b = subnet_obj.Subnet(
            self.context,
            id=self.ext_sub_b_id,
            project_id=self.tenant_id,
            ip_version=constants.IP_VERSION_4,
            cidr=net_utils.AuthenticIPNetwork('192.0.2.128/25'),
            gateway_ip=netaddr.IPAddress('192.0.2.129'),
            network_id=self.ext_net_b_id)
        self.ext_sub_b.create()

        self.ext_net_c_id = _uuid()
        self.ext_sub_c_id = _uuid()
        self.network_c = net_obj.Network(
            self.context,
            id=self.ext_net_c_id,
            project_id=self.tenant_id,
            admin_state_up=True,
            status=constants.NET_STATUS_ACTIVE)
        self.network_c.create()
        self.net_ext_c = net_obj.ExternalNetwork(
            self.context, network_id=self.ext_net_c_id)
        self.net_ext_c.create()

        self.ext_sub_c = subnet_obj.Subnet(
            self.context,
            id=self.ext_sub_c_id,
            project_id=self.tenant_id,
            ip_version=constants.IP_VERSION_4,
            # Overlaps with subnet A above on purpose for overlap testing.
            cidr=net_utils.AuthenticIPNetwork('192.0.2.0/25'),
            gateway_ip=netaddr.IPAddress('192.0.2.1'),
            network_id=self.ext_net_c_id)
        self.ext_sub_c.create()

        # Create an IPAM subnet for fixed ip allocations.
        self.ipam_ext_subnet_a = ipam_obj.IpamSubnet(
            self.context,
            neutron_subnet_id=self.ext_sub_a_id)
        self.ipam_ext_subnet_a.create()

        self.ipam_ext_subnet_b = ipam_obj.IpamSubnet(
            self.context,
            neutron_subnet_id=self.ext_sub_b_id)
        self.ipam_ext_subnet_b.create()

        self.ipam_ext_subnet_c = ipam_obj.IpamSubnet(
            self.context,
            neutron_subnet_id=self.ext_sub_c_id)
        self.ipam_ext_subnet_c.create()

        # Create an allocation pool that will use the IPAM subnet.
        self.ipam_ext_pool_a = ipam_obj.IpamAllocationPool(
            self.context,
            id=_uuid(),
            ipam_subnet_id=self.ipam_ext_subnet_a.id,
            first_ip='192.0.2.3',
            last_ip='192.0.2.126',
        )
        self.ipam_ext_pool_a.create()

        self.ipam_ext_pool_b = ipam_obj.IpamAllocationPool(
            self.context,
            id=_uuid(),
            ipam_subnet_id=self.ipam_ext_subnet_b.id,
            first_ip='192.0.2.131',
            last_ip='192.0.2.254',
        )
        self.ipam_ext_pool_b.create()

        self.ipam_ext_pool_c = ipam_obj.IpamAllocationPool(
            self.context,
            id=_uuid(),
            ipam_subnet_id=self.ipam_ext_subnet_c.id,
            first_ip='192.0.2.3',
            last_ip='192.0.2.126',
        )
        self.ipam_ext_pool_c.create()

        # Create a router that will be modified during the tests.
        self.router = l3_models.Router(
            id=_uuid(),
            name=None,
            tenant_id=self.tenant_id,
            admin_state_up=True,
            status=constants.NET_STATUS_ACTIVE,
            enable_snat=True,
            gw_port_id=None)

        self.context.session.add(self.router)
        self.context.session.expire_all()
        self.context.session.commit()

    def test_add_external_gateways_trivial(self):
        ext_gws = []
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        # A trivial case with an empty list passed in.
        result = self.target_object._add_external_gateways(
            self.context, self.router.id, ext_gws, body)
        self.assertEqual([], result)

    def test_add_external_gateways_single(self):
        ext_gws = [{"network_id": self.ext_net_a_id}]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        result = self.target_object.add_external_gateways(
            self.context, self.router.id, body)

        res_gw_a = result['router']['external_gateways'][0]

        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)
        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        gw_ports = port_obj.Port.get_ports_by_router_and_network(
            self.context, self.router.id, constants.DEVICE_OWNER_ROUTER_GW,
            self.ext_net_a_id)
        self.assertEqual(len(gw_ports), 1)

        gw_port = gw_ports[0]
        self.assertEqual(new_router['gw_port_id'], gw_port['id'])

    def test_add_external_gateways_multiple(self):
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        result = self.target_object.add_external_gateways(
            self.context, self.router.id, body)

        res_gw_a = result['router']['external_gateways'][0]
        res_gw_b = result['router']['external_gateways'][1]

        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertEqual(res_gw_b['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])
        self.assertIsNotNone(res_gw_b['external_fixed_ips'])

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)

        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})

        self.assertEqual(len(gw_ports), 2)

        # Now check that calling the ADD API multiple times succeeds.
        ext_gws = [
            {"network_id": self.ext_net_b_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        result = self.target_object.add_external_gateways(
            self.context, self.router.id, body)

        res_gw_a = result['router']['external_gateways'][0]
        res_gw_b = result['router']['external_gateways'][1]
        res_gw_c = result['router']['external_gateways'][2]

        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertEqual(res_gw_b['network_id'], self.ext_net_b_id)
        self.assertEqual(res_gw_c['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])
        self.assertIsNotNone(res_gw_b['external_fixed_ips'])
        self.assertIsNotNone(res_gw_c['external_fixed_ips'])

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)

        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})

        self.assertEqual(len(gw_ports), 3)

        # Check that adding a gateway with already allocated fixed IPs fails.
        ext_gws = [
            {"network_id": self.ext_net_b_id,
             "external_fixed_ips": res_gw_c['external_fixed_ips']},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        self.assertRaises(
            ipam_exceptions.IpAddressAlreadyAllocated,
            self.target_object.add_external_gateways, self.context,
            self.router.id, body
        )

    def test_remove_external_gateways_trivial(self):
        ext_gws = []
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        # A trivial case with an empty list passed in.
        result = self.target_object.remove_external_gateways(
            self.context, self.router.id, body)
        self.assertIsNone(result['router']['external_gateway_info'])

    def test_remove_external_gateways_single(self):
        ext_gws = [{"network_id": self.ext_net_a_id}]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        self.target_object.add_external_gateways(
            self.context, self.router.id, body)
        self.assertIsNotNone(self.router.gw_port_id)

        result = self.target_object.remove_external_gateways(
            self.context, self.router.id, body)
        self.assertIsNone(self.router.gw_port_id)
        self.assertIsNone(result['router']['external_gateway_info'])

    def test_remove_external_gateways_multiple(self):
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        self.target_object.add_external_gateways(
            self.context, self.router.id, body)
        self.assertIsNotNone(self.router.gw_port_id)

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})
        self.assertEqual(len(gw_ports), 2)

        result = self.target_object.remove_external_gateways(
            self.context, self.router.id, body)
        self.assertIsNone(self.router.gw_port_id)
        self.assertIsNone(result['router']['external_gateway_info'])

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})
        self.assertEqual(len(gw_ports), 0)

    def test_remove_external_gateways_remove_compat(self):
        '''Test removal of a compatibility gateway port using the new API.

        When removing a compatibility gateway port using the new API we need
        to make sure that an existing extra gateway port takes it place instead
        as a compatibility gateway port.
        '''
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        add_body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        remove_body = {
            "router": {
                "external_gateways": ext_gws[:1]
            }
        }

        self.target_object.add_external_gateways(
            self.context, self.router.id, add_body)
        self.assertIsNotNone(self.router.gw_port_id)

        old_gw_port_id = self.router.gw_port_id

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})
        self.assertEqual(len(gw_ports), 2)

        self.target_object.remove_external_gateways(
            self.context, self.router.id, remove_body)

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})
        self.assertEqual(len(gw_ports), 1)

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)
        self.assertNotEqual(old_gw_port_id, new_router['gw_port_id'])
        self.assertEqual(new_router['external_gateway_info']['network_id'],
                         self.ext_net_b_id)

    def test_update_external_gateways_add_pristine_and_remove(self):
        '''Test the addition of external gateway ports using the update API.'''
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        add_body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        result = self.target_object.update_external_gateways(
            self.context, self.router.id, add_body)
        self.assertIsNotNone(self.router.gw_port_id)

        res_gw_a = result['router']['external_gateways'][0]
        res_gw_b = result['router']['external_gateways'][1]
        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertEqual(res_gw_b['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])
        self.assertIsNotNone(res_gw_b['external_fixed_ips'])

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)

        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})

        self.assertEqual(len(gw_ports), 2)

        # Reorder gateways.
        ext_gws = [
            {"network_id": self.ext_net_b_id},
            {"network_id": self.ext_net_a_id},
        ]
        update_body = {
            "router": {
                "external_gateways": ext_gws
            }
        }
        result = self.target_object.update_external_gateways(
            self.context, self.router.id, update_body)
        self.assertIsNotNone(self.router.gw_port_id)

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)
        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        # The compat gateway should now have a different network_id.
        res_gw_b = result['router']['external_gateways'][0]
        res_gw_a = result['router']['external_gateways'][1]
        self.assertEqual(res_gw_b['network_id'], self.ext_net_b_id)
        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])
        self.assertIsNotNone(res_gw_b['external_fixed_ips'])

        # Remove one gateway.
        update_body = {
            "router": {
                "external_gateways": ext_gws[1:]
            }
        }

        result = self.target_object.update_external_gateways(
            self.context, self.router.id, update_body)
        self.assertIsNotNone(self.router.gw_port_id)

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)
        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        res_gw_a = result['router']['external_gateways'][0]
        self.assertEqual(res_gw_a['network_id'], self.ext_net_a_id)
        self.assertIsNotNone(res_gw_a['external_fixed_ips'])

        # Clear all gateways.
        update_body = {
            "router": {
                "external_gateways": {}
            }
        }

        result = self.target_object.update_external_gateways(
            self.context, self.router.id, update_body)
        self.assertIsNone(self.router.gw_port_id)

    def test_compat_remove_via_update(self):
        '''Test the removal of a gateway port using the compat API.

        Removal of a compat gateway in the presence of an extra
        gateway port should make that extra gateway port a compat
        gateway port.
        '''
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        result = self.target_object.add_external_gateways(
            self.context, self.router.id, body)

        update_body = {
            "router": {
                "external_gateway_info": {}
            }
        }
        # Now perform an update with an empty gw info to remove the current
        # compat gw port.
        result = self.target_object.update_router(self.context, self.router.id,
                                                  update_body)

        # The existing extra gateway port should now take its place.
        res_gw = result['external_gateways'][0]
        self.assertEqual(res_gw['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(res_gw['external_fixed_ips'])
        self.assertEqual(len(result['external_gateways']), 1)

        new_router = self.target_object.get_router(self.context,
                                                   self.router.id)

        new_gw_info = new_router['external_gateway_info']
        self.assertEqual(new_gw_info['network_id'], self.ext_net_b_id)
        self.assertIsNotNone(new_gw_info['external_fixed_ips'])

        gw_ports = l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})

        self.assertEqual(len(gw_ports), 1)

    def test_update_fixed_ip(self):
        '''Test updating a fixed IP of an existing port.'''
        ext_gws = [
            {"network_id": self.ext_net_a_id},
            {"network_id": self.ext_net_b_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        result = self.target_object.add_external_gateways(
            self.context, self.router.id, body)

        gw_ports_initial = [o.port_id for o in l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})]

        fips = copy.deepcopy(
            result['router']['external_gateways'][1]['external_fixed_ips'])
        # Append a fixed ip not used in the allocation pool. The existing
        # one should be used to find an existing port.
        fips.append({'ip_address': '192.0.2.130',
                     'subnet_id': fips[0]['subnet_id']})
        expected_fips = copy.deepcopy(fips)

        update_body = {
            "router": {
                "external_gateways": [
                    {"network_id": self.ext_net_a_id},
                    # Use the new set of fixed IPs in the request.
                    {"network_id": self.ext_net_b_id,
                     "external_fixed_ips": fips},
                ]}
        }
        result = self.target_object.update_external_gateways(
            self.context, self.router.id,
            update_body)

        self.assertCountEqual(
            result['router']['external_gateways'][1]['external_fixed_ips'],
            expected_fips,
        )

        gw_ports_final = [o.port_id for o in l3_obj.RouterPort.get_objects(
            self.context,
            **{'router_id': self.router.id,
               'port_type': constants.DEVICE_OWNER_ROUTER_GW})]

        # Make sure the ports are not recreated in the process, i.e. port IDs
        # stay the same.
        self.assertCountEqual(
            gw_ports_initial,
            gw_ports_final,
        )

    def test_add_external_gateways_overlapping_subnets(self):
        ext_gws = [
            {"network_id": self.ext_net_a_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        self.target_object.add_external_gateways(self.context, self.router.id,
                                                 body)

        ext_gws = [
            {"network_id": self.ext_net_c_id},
        ]
        body = {
            "router": {
                "external_gateways": ext_gws
            }
        }

        self.assertRaisesRegex(
            exceptions.BadRequest,
            'Bad router request: Cidr 192.0.2.0/25 of subnet'
            f' {self.ext_sub_c_id} overlaps with cidr 192.0.2.0/25 of '
            f'subnet {self.ext_sub_a_id}.',
            self.target_object.add_external_gateways, self.context,
            self.router.id, body
        )



# Copyright (c) 2023 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.conf import service
from neutron.tests import base as tests_base


class GetRpcWorkers(tests_base.BaseTestCase):

    def test_no_previous_registration(self):
        self.assertIsNone(service.get_rpc_workers(conf=cfg.CONF))
        cfg.CONF.set_override('rpc_workers', 150)
        self.assertEqual(150, service.get_rpc_workers(conf=cfg.CONF))

    def test_previous_registration(self):
        service.register_service_opts(service.SERVICE_OPTS, conf=cfg.CONF)
        self.assertIsNone(service.get_rpc_workers(conf=cfg.CONF))
        cfg.CONF.set_override('rpc_workers', 200)
        self.assertEqual(200, service.get_rpc_workers(conf=cfg.CONF))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron_lib import context
from neutron_lib.db import model_query
from neutron_lib import exceptions as n_exc
from neutron_lib.objects import utils as obj_utils

from neutron.objects import base
from neutron.objects.db import api
from neutron.objects import network
from neutron.tests import base as test_base
from neutron.tests.unit import testlib_api


class FakeModel(object):
    def __init__(self, *args, **kwargs):
        pass


class FakeObj(base.NeutronDbObject):
    db_model = FakeModel


class GetObjectsTestCase(test_base.BaseTestCase):

    def test_get_objects_pass_marker_obj_when_limit_and_marker_passed(self):
        ctxt = context.get_admin_context()
        marker = mock.sentinel.marker
        limit = mock.sentinel.limit
        pager = base.Pager(marker=marker, limit=limit)

        with mock.patch.object(
                model_query, 'get_collection') as get_collection:
            with mock.patch.object(api, 'get_object') as get_object:
                api.get_objects(FakeObj, ctxt, _pager=pager)
        get_object.assert_called_with(FakeObj, ctxt, id=marker)
        get_collection.assert_called_with(
            ctxt, FakeObj.db_model, dict_func=None,
            filters={},
            limit=limit,
            marker_obj=get_object.return_value)


class GetValuesTestCase(test_base.BaseTestCase):

    def test_get_values(self):
        ctxt = context.get_admin_context()
        fake_field = 'fake_field'

        with mock.patch.object(
                model_query, 'get_values') as get_values:
            api.get_values(FakeObj, ctxt, fake_field)
        get_values.assert_called_with(
            ctxt, FakeObj.db_model, fake_field, filters={})


class CreateObjectTestCase(test_base.BaseTestCase):
    def test_populate_id(self, populate_id=True):
        ctxt = context.get_admin_context()
        values = {'x': 1, 'y': 2, 'z': 3}
        with mock.patch.object(FakeObj, 'db_model') as db_model_mock:
            with mock.patch.object(ctxt.__class__, 'session'):
                api.create_object(FakeObj, ctxt, values,
                                  populate_id=populate_id)
        expected = copy.copy(values)
        if populate_id:
            expected['id'] = mock.ANY
        db_model_mock.assert_called_with(**expected)

    def test_populate_id_False(self):
        self.test_populate_id(populate_id=False)


class CRUDScenarioTestCase(testlib_api.SqlTestCase):

    CORE_PLUGIN = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'

    def setUp(self):
        super(CRUDScenarioTestCase, self).setUp()
        # TODO(ihrachys): revisit plugin setup once we decouple
        # neutron.objects.db.api from core plugin instance
        self.setup_coreplugin(self.CORE_PLUGIN)
        # NOTE(ihrachys): nothing specific to networks in this test case, but
        # we needed to pick some real object, so we picked the network. Any
        # other object would work as well for our needs here.
        self.obj_cls = network.Network
        self.ctxt = context.get_admin_context()

    def _compare_objs(self, obj1, obj2):
        for field in (field for field in self.obj_cls.fields if
                      field not in ('updated_at', 'created_at')):
            self.assertEqual(getattr(obj1, field, None),
                             getattr(obj2, field, None))

    def test_get_object_with_None_value_in_filters(self):
        obj = api.create_object(self.obj_cls, self.ctxt, {'name': 'foo'})
        new_obj = api.get_object(
            self.obj_cls, self.ctxt, name='foo', status=None)
        self._compare_objs(obj, new_obj)

    def test_get_objects_with_None_value_in_filters(self):
        obj = api.create_object(self.obj_cls, self.ctxt, {'name': 'foo'})
        new_objs = api.get_objects(
            self.obj_cls, self.ctxt, name='foo', status=None)
        self._compare_objs(obj, new_objs[0])

    def test_get_objects_with_string_matching_filters_contains(self):
        obj1 = api.create_object(
            self.obj_cls, self.ctxt, {'name': 'obj_con_1'})
        obj2 = api.create_object(
            self.obj_cls, self.ctxt, {'name': 'obj_con_2'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj_3'})

        objs = api.get_objects(
            self.obj_cls, self.ctxt, name=obj_utils.StringContains('con'),
            _pager=base.Pager(sorts=[('name', True)]))
        self.assertEqual(2, len(objs))
        self._compare_objs(obj1, objs[0])
        self._compare_objs(obj2, objs[1])

    def test_get_objects_with_string_matching_filters_starts(self):
        obj1 = api.create_object(self.obj_cls, self.ctxt, {'name': 'pre_obj1'})
        obj2 = api.create_object(self.obj_cls, self.ctxt, {'name': 'pre_obj2'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj_3'})

        objs = api.get_objects(
            self.obj_cls, self.ctxt, name=obj_utils.StringStarts('pre'),
            _pager=base.Pager(sorts=[('name', True)]))
        self.assertEqual(2, len(objs))
        self._compare_objs(obj1, objs[0])
        self._compare_objs(obj2, objs[1])

    def test_get_objects_with_string_matching_filters_ends(self):
        obj1 = api.create_object(self.obj_cls, self.ctxt, {'name': 'obj1_end'})
        obj2 = api.create_object(self.obj_cls, self.ctxt, {'name': 'obj2_end'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj_3'})

        objs = api.get_objects(
            self.obj_cls, self.ctxt, name=obj_utils.StringEnds('end'),
            _pager=base.Pager(sorts=[('name', True)]))
        self.assertEqual(2, len(objs))
        self._compare_objs(obj1, objs[0])
        self._compare_objs(obj2, objs[1])

    def test_get_values_with_None_value_in_filters(self):
        api.create_object(self.obj_cls, self.ctxt, {'name': 'foo'})
        values = api.get_values(
            self.obj_cls, self.ctxt, 'name', name='foo', status=None)
        self.assertEqual('foo', values[0])

    def test_get_values_with_string_matching_filters_contains(self):
        api.create_object(
            self.obj_cls, self.ctxt, {'name': 'obj_con_1'})
        api.create_object(
            self.obj_cls, self.ctxt, {'name': 'obj_con_2'})
        api.create_object(
            self.obj_cls, self.ctxt, {'name': 'obj_3'})

        values = api.get_values(
            self.obj_cls, self.ctxt, 'name',
            name=obj_utils.StringContains('con'))
        self.assertEqual(2, len(values))
        self.assertIn('obj_con_1', values)
        self.assertIn('obj_con_2', values)
        self.assertNotIn('obj_3', values)

    def test_get_values_with_string_matching_filters_starts(self):
        api.create_object(self.obj_cls, self.ctxt, {'name': 'pre_obj1'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'pre_obj2'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj_3'})

        values = api.get_values(
            self.obj_cls, self.ctxt, 'name',
            name=obj_utils.StringStarts('pre'))
        self.assertEqual(2, len(values))
        self.assertIn('pre_obj1', values)
        self.assertIn('pre_obj2', values)
        self.assertNotIn('obj_3', values)

    def test_get_values_with_string_matching_filters_ends(self):
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj1_end'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj2_end'})
        api.create_object(self.obj_cls, self.ctxt, {'name': 'obj_3'})

        values = api.get_values(
            self.obj_cls, self.ctxt, 'name', name=obj_utils.StringEnds('end'))
        self.assertEqual(2, len(values))
        self.assertIn('obj1_end', values)
        self.assertIn('obj2_end', values)
        self.assertNotIn('obj_3', values)

    def test_get_object_create_update_delete(self):
        obj = api.create_object(self.obj_cls, self.ctxt, {'name': 'foo'})

        new_obj = api.get_object(self.obj_cls, self.ctxt, id=obj.id)
        self._compare_objs(obj, new_obj)

        obj.name = 'bar'
        api.update_object(self.obj_cls, self.ctxt, {'name': 'bar'}, id=obj.id)

        new_obj = api.get_object(self.obj_cls, self.ctxt, id=obj.id)
        self._compare_objs(obj, new_obj)

        api.delete_object(self.obj_cls, self.ctxt, id=obj.id)

        new_obj = api.get_object(self.obj_cls, self.ctxt, id=obj.id)
        self.assertIsNone(new_obj)

        # delete_object raises an exception on missing object
        self.assertRaises(
            n_exc.ObjectNotFound,
            api.delete_object, self.obj_cls, self.ctxt, id=obj.id)

        # but delete_objects does not not
        api.delete_objects(self.obj_cls, self.ctxt, id=obj.id)

    def test_delete_objects_removes_all_matching_objects(self):
        # create some objects with identical description
        for i in range(10):
            api.create_object(
                self.obj_cls, self.ctxt,
                {'name': 'foo%d' % i, 'description': 'bar'})
        # create some more objects with a different description
        descriptions = set()
        for i in range(10, 20):
            desc = 'bar%d' % i
            descriptions.add(desc)
            api.create_object(
                self.obj_cls, self.ctxt,
                {'name': 'foo%d' % i, 'description': desc})
        # make sure that all objects are in the database
        self.assertEqual(20, api.count(self.obj_cls, self.ctxt))
        # now delete just those with the 'bar' description
        api.delete_objects(self.obj_cls, self.ctxt, description='bar')

        # check that half of objects are gone, and remaining have expected
        # descriptions
        objs = api.get_objects(self.obj_cls, self.ctxt)
        self.assertEqual(10, len(objs))
        self.assertEqual(
            descriptions,
            {obj.description for obj in objs})



# Copyright (c) 2022 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg

from neutron.agent.l2.extensions.metadata import host_metadata_proxy
from neutron.agent.linux import external_process
from neutron.tests import base


class TestHostMedataHAProxyDaemonMonitor(base.BaseTestCase):

    def setUp(self):
        super(TestHostMedataHAProxyDaemonMonitor, self).setUp()

        self.ensure_dir = mock.patch(
            'oslo_utils.fileutils.ensure_tree').start()

        self.utils_exec_p = mock.patch(
            'neutron.agent.linux.utils.execute')
        self.utils_exec = self.utils_exec_p.start()

        self.delete_if_exists = mock.patch(
            'neutron.agent.linux.utils.delete_if_exists').start()

        self.utils_replace_file_p = mock.patch(
            'neutron_lib.utils.file.replace_file')
        self.utils_replace_file = self.utils_replace_file_p.start()

    def test_spawn_host_metadata_haproxy(self):
        cfg.CONF.set_override('metadata_proxy_shared_secret',
                              'secret', group='METADATA')
        conffile = '/fake/host_metadata_proxy.haproxy.conf'
        pidfile = '/fake/host_metadata_proxy.pid.haproxy'
        process_monitor = external_process.ProcessMonitor(
            config=cfg.CONF,
            resource_type='MetadataPath')

        get_conf_file_name = 'neutron.agent.linux.utils.get_conf_file_name'
        get_pid_file_name = ('neutron.agent.linux.external_process.'
                             'ProcessManager.get_pid_file_name')
        utils_execute = 'neutron.agent.common.utils.execute'

        mock.patch(get_conf_file_name).start().return_value = conffile
        mock.patch(get_pid_file_name).start().return_value = pidfile
        execute = mock.patch(utils_execute).start()

        host_meta = host_metadata_proxy.HostMedataHAProxyDaemonMonitor(
            process_monitor)
        instance_infos = [
            {"instance_id": "uuid1",
             "provider_ip": "1.1.1.1",
             "project_id": "project1"}]
        host_meta.config(instance_infos)
        host_meta.enable()
        cmd = execute.call_args[0][0]
        _join = lambda *args: ' '.join(args)
        cmd = _join(*cmd)
        self.assertIn('haproxy', cmd)
        self.assertIn(_join('-f', conffile), cmd)
        self.assertIn(_join('-p', pidfile), cmd)
        self.delete_if_exists.assert_called_once_with(pidfile,
                                                      run_as_root=True)

    def test_generate_host_metadata_haproxy_config(self):
        cfg.CONF.set_override('metadata_proxy_shared_secret',
                              'secret', group='METADATA')
        sig = (
            "3b5421875d7ba0fc910202f5ce448d9419597e7b66f702b53335116fee60e81e")
        cfg.CONF.set_override('nova_metadata_host',
                              '2.2.2.2',
                              group='METADATA')
        cfg.CONF.set_override('nova_metadata_port',
                              '8775',
                              group='METADATA')
        process_monitor = external_process.ProcessMonitor(
            config=cfg.CONF,
            resource_type='MetadataPath')
        host_meta = host_metadata_proxy.HostMedataHAProxyDaemonMonitor(
            process_monitor)
        instance_infos = [
            host_metadata_proxy.ProxyInstance('uuid1', '1.1.1.1', 'project1')]
        host_meta._generate_proxy_conf(instance_infos)
        acl = "acl instance_uuid1_1.1.1.1 src 1.1.1.1"
        use_acl = "use_backend backend_uuid1_1.1.1.1 if instance_uuid1_1.1.1.1"
        backend = "backend backend_uuid1_1.1.1.1"
        http_hd_ins_id = "http-request set-header X-Instance-ID uuid1"
        http_hd_pj = "http-request set-header X-Tenant-ID project1"
        http_hd_sig = (
            "http-request set-header X-Instance-ID-Signature %s" % sig)
        meta_real_srv = "server metasrv 2.2.2.2:8775"
        expects = [acl, use_acl, backend, http_hd_ins_id, http_hd_pj,
                   http_hd_sig, meta_real_srv]
        for exp in expects:
            self.assertIn(exp, self.utils_replace_file.call_args[0][1])



# Copyright (c) 2014 OpenStack Foundation, all rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from unittest import mock
import warnings

import netaddr
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants
from neutron_lib import context
from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from oslo_utils import uuidutils
from sqlalchemy.orm import exc
from sqlalchemy.orm import query

from neutron.db import db_base_plugin_v2
from neutron.db.models import l3 as l3_models
from neutron.db import segments_db
from neutron.objects import network as network_obj
from neutron.objects import ports as port_obj
from neutron.plugins.ml2 import db as ml2_db
from neutron.plugins.ml2 import models
from neutron.services.segments import exceptions as seg_exc
from neutron.tests.unit import testlib_api


PLUGIN_NAME = 'ml2'


class Ml2DBTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(Ml2DBTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        self.setup_coreplugin(PLUGIN_NAME)

    def _setup_neutron_network(self, network_id):
        network_obj.Network(self.ctx, id=network_id).create()

    def _setup_neutron_port(self, network_id, port_id):
        mac_address = db_base_plugin_v2.NeutronDbPluginV2._generate_macs()[0]
        port = port_obj.Port(self.ctx,
                             id=port_id,
                             network_id=network_id,
                             mac_address=netaddr.EUI(mac_address),
                             admin_state_up=True,
                             status='DOWN',
                             device_id='',
                             device_owner='')
        port.create()
        return port

    def _setup_neutron_portbinding(self, port_id, vif_type, host,
                                   status=constants.ACTIVE):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            self.ctx.session.add(models.PortBinding(port_id=port_id,
                                                    vif_type=vif_type,
                                                    host=host,
                                                    status=status))

    @staticmethod
    def _sort_segments(segments):
        return sorted(segments, key=lambda d: d['segmentation_id'])

    def _create_segments(self, segments, is_seg_dynamic=False,
                         network_id=uuidutils.generate_uuid()):
        self._setup_neutron_network(network_id)
        for segment in segments:
            segments_db.add_network_segment(
                self.ctx, network_id, segment,
                is_dynamic=is_seg_dynamic)
            segment['network_id'] = network_id

        net_segments = segments_db.get_network_segments(
                           self.ctx, network_id,
                           filter_dynamic=is_seg_dynamic)
        net_segments = self._sort_segments(net_segments)

        for segment_index, segment in enumerate(segments):
            self.assertEqual(segment, net_segments[segment_index])

        return net_segments

    def test_network_segments_for_provider_network(self):
        segment = {api.NETWORK_TYPE: 'vlan',
                   api.PHYSICAL_NETWORK: 'physnet1',
                   api.SEGMENTATION_ID: 1}
        self._create_segments([segment])

    def test_network_segments_is_dynamic_true(self):
        segment = {api.NETWORK_TYPE: 'vlan',
                   api.PHYSICAL_NETWORK: 'physnet1',
                   api.SEGMENTATION_ID: 1}
        self._create_segments([segment], is_seg_dynamic=True)

    def test_network_segments_for_multiprovider_network(self):
        segments = [{api.NETWORK_TYPE: 'vlan',
                    api.PHYSICAL_NETWORK: 'physnet1',
                    api.SEGMENTATION_ID: 1},
                    {api.NETWORK_TYPE: 'vlan',
                     api.PHYSICAL_NETWORK: 'physnet2',
                     api.SEGMENTATION_ID: 2}]
        self._create_segments(segments)

    def test_get_networks_segments(self):
        net_id1 = uuidutils.generate_uuid()
        net_id2 = uuidutils.generate_uuid()
        segments1 = [{api.NETWORK_TYPE: 'vlan',
                      api.PHYSICAL_NETWORK: 'physnet1',
                      api.SEGMENTATION_ID: 1},
                     {api.NETWORK_TYPE: 'vlan',
                      api.PHYSICAL_NETWORK: 'physnet2',
                      api.SEGMENTATION_ID: 2}]
        segments2 = [{api.NETWORK_TYPE: 'vlan',
                      api.PHYSICAL_NETWORK: 'physnet1',
                      api.SEGMENTATION_ID: 3},
                     {api.NETWORK_TYPE: 'vlan',
                      api.PHYSICAL_NETWORK: 'physnet2',
                      api.SEGMENTATION_ID: 4}]
        net1segs = self._create_segments(segments1, network_id=net_id1)
        net2segs = self._create_segments(segments2, network_id=net_id2)
        segs = segments_db.get_networks_segments(
            self.ctx, [net_id1, net_id2])
        self.assertEqual(net1segs, self._sort_segments(segs[net_id1]))
        self.assertEqual(net2segs, self._sort_segments(segs[net_id2]))

    def test_get_networks_segments_no_segments(self):
        net_id1 = uuidutils.generate_uuid()
        net_id2 = uuidutils.generate_uuid()
        self._create_segments([], network_id=net_id1)
        self._create_segments([], network_id=net_id2)
        segs = segments_db.get_networks_segments(
            self.ctx, [net_id1, net_id2])
        self.assertEqual([], segs[net_id1])
        self.assertEqual([], segs[net_id2])

    def test_get_segment_by_id(self):
        segment = {api.NETWORK_TYPE: 'vlan',
                   api.PHYSICAL_NETWORK: 'physnet1',
                   api.SEGMENTATION_ID: 1}

        net_segment = self._create_segments([segment])[0]
        segment_uuid = net_segment[api.ID]

        net_segment = segments_db.get_segment_by_id(self.ctx,
                                                    segment_uuid)
        self.assertEqual(segment, net_segment)

    def test_get_segment_by_id_result_not_found(self):
        segment_uuid = uuidutils.generate_uuid()
        net_segment = segments_db.get_segment_by_id(self.ctx,
                                                    segment_uuid)
        self.assertIsNone(net_segment)

    def test_delete_network_segment(self):
        segment = {api.NETWORK_TYPE: 'vlan',
                   api.PHYSICAL_NETWORK: 'physnet1',
                   api.SEGMENTATION_ID: 1}

        net_segment = self._create_segments([segment])[0]
        segment_uuid = net_segment[api.ID]

        segments_db.delete_network_segment(self.ctx, segment_uuid)
        # Get segment and verify its empty
        net_segment = segments_db.get_segment_by_id(self.ctx,
                                                    segment_uuid)
        self.assertIsNone(net_segment)

    def test_get_dynamic_segment(self):
        net_id = uuidutils.generate_uuid()
        segment1 = {api.NETWORK_TYPE: 'vlan',
                    api.PHYSICAL_NETWORK: 'physnet1',
                    api.SEGMENTATION_ID: 1}

        self._create_segments(
            [segment1], is_seg_dynamic=True, network_id=net_id)

        segs1 = segments_db.get_dynamic_segment(
            self.ctx, net_id)
        self.assertEqual('vlan', segs1[api.NETWORK_TYPE])
        self.assertEqual('physnet1', segs1[api.PHYSICAL_NETWORK])
        self.assertEqual(1, segs1[api.SEGMENTATION_ID])

        segs2 = segments_db.get_dynamic_segment(
            self.ctx, net_id, physical_network='physnet1')
        self.assertEqual('vlan', segs2[api.NETWORK_TYPE])
        self.assertEqual('physnet1', segs2[api.PHYSICAL_NETWORK])
        self.assertEqual(1, segs2[api.SEGMENTATION_ID])

        segs3 = segments_db.get_dynamic_segment(
            self.ctx, net_id, segmentation_id=1)
        self.assertEqual('vlan', segs3[api.NETWORK_TYPE])
        self.assertEqual('physnet1', segs3[api.PHYSICAL_NETWORK])
        self.assertEqual(1, segs3[api.SEGMENTATION_ID])

    def test_add_port_binding(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        self._setup_neutron_network(network_id)
        self._setup_neutron_port(network_id, port_id)

        port = ml2_db.add_port_binding(self.ctx, port_id)
        self.assertEqual(port_id, port.port_id)
        self.assertEqual(portbindings.VIF_TYPE_UNBOUND, port.vif_type)

    def test_get_port_binding_host(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        host = 'fake_host'
        other_host = 'other_fake_host'
        vif_type = portbindings.VIF_TYPE_UNBOUND
        self._setup_neutron_network(network_id)
        self._setup_neutron_port(network_id, port_id)
        self._setup_neutron_portbinding(port_id, vif_type, host)
        self._setup_neutron_portbinding(port_id, vif_type, other_host,
                                        status=constants.INACTIVE)

        port_host = ml2_db.get_port_binding_host(self.ctx, port_id)
        self.assertEqual(host, port_host)

    def test_get_port_binding_host_multiple_results_found(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        port_id_one = uuidutils.generate_uuid()
        port_id_two = uuidutils.generate_uuid()
        # NOTE(manjeets) to check startswith testcase we
        # need port ids with same prefix
        port_id_one = port_id[:8] + port_id_one[8:]
        port_id_two = port_id[:8] + port_id_two[8:]
        host = 'fake_host'
        vif_type = portbindings.VIF_TYPE_UNBOUND
        self._setup_neutron_network(network_id)
        self._setup_neutron_port(network_id, port_id_one)
        self._setup_neutron_portbinding(port_id_one, vif_type, host)
        self._setup_neutron_port(network_id, port_id_two)
        self._setup_neutron_portbinding(port_id_two, vif_type, host)

        port_host = ml2_db.get_port_binding_host(self.ctx, port_id[:8])
        self.assertIsNone(port_host)

    def test_get_port_binding_host_result_not_found(self):
        port_id = uuidutils.generate_uuid()

        port_host = ml2_db.get_port_binding_host(self.ctx, port_id)
        self.assertIsNone(port_host)

    def test_get_port(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        self._setup_neutron_network(network_id)
        self._setup_neutron_port(network_id, port_id)

        port = ml2_db.get_port(self.ctx, port_id)
        self.assertEqual(port_id, port.id)

    def test_get_port_multiple_results_found(self):
        with mock.patch(
                'sqlalchemy.orm.query.Query.one',
                side_effect=exc.MultipleResultsFound):
            port = ml2_db.get_port(self.ctx, 'unused')
        self.assertIsNone(port)

    def test_get_port_result_not_found(self):
        port_id = uuidutils.generate_uuid()
        port = ml2_db.get_port(self.ctx, port_id)
        self.assertIsNone(port)

    def test_generating_multiple_mac_addresses(self):
        mac_regex = "^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"

        macs = db_base_plugin_v2.NeutronDbPluginV2._generate_macs()
        self.assertEqual(1, len(macs))
        self.assertIsNotNone(re.search(mac_regex, macs[0]))

        macs = db_base_plugin_v2.NeutronDbPluginV2._generate_macs(5)
        self.assertEqual(5, len(macs))
        for mac in macs:
            self.assertIsNotNone(re.search(mac_regex, mac))

    def test__prevent_segment_delete_with_port_bound_raise(self):
        payload_mock = mock.Mock()
        payload_mock.metadata.get.return_value = False
        payload_mock.context = self.ctx
        with mock.patch.object(
                 port_obj.Port,
                 'get_auto_deletable_port_ids_and_proper_port_count_by_segment'
             ) as mock_get:
            mock_get.return_value = ([], 1)
            self.assertRaises(
                seg_exc.SegmentInUse,
                ml2_db._prevent_segment_delete_with_port_bound,
                resource=mock.Mock(),
                event=mock.Mock(),
                trigger=mock.Mock(),
                payload=payload_mock,
            )

    def test__prevent_segment_delete_with_port_bound_auto_delete(self):
        payload_mock = mock.Mock()
        payload_mock.metadata.get.return_value = False
        payload_mock.context = self.ctx
        plugin = directory.get_plugin()
        with mock.patch.object(
                 port_obj.Port,
                 'get_auto_deletable_port_ids_and_proper_port_count_by_segment'
             ) as mock_get, \
                mock.patch.object(plugin, 'delete_port') as mock_delete_port:
            mock_get.return_value = (['fake-port'], 0)
            ml2_db._prevent_segment_delete_with_port_bound(
                resource=mock.Mock(),
                event=mock.Mock(),
                trigger=mock.Mock(),
                payload=payload_mock,
            )
            mock_delete_port.assert_called_with(mock.ANY, 'fake-port')


class Ml2DvrDBTestCase(testlib_api.SqlTestCase):

    def setUp(self):
        super(Ml2DvrDBTestCase, self).setUp()
        self.ctx = context.get_admin_context()
        self.setup_coreplugin(PLUGIN_NAME)

    def _setup_neutron_network(self, network_id, port_ids):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            network_obj.Network(self.ctx, id=network_id).create()
            ports = []
            for port_id in port_ids:
                mac_address = (db_base_plugin_v2.NeutronDbPluginV2.
                               _generate_macs()[0])
                port = port_obj.Port(self.ctx,
                                     id=port_id,
                                     network_id=network_id,
                                     mac_address=netaddr.EUI(mac_address),
                                     admin_state_up=True,
                                     status='ACTIVE',
                                     device_id='',
                                     device_owner='')
                port.create()
                ports.append(port)
            return ports

    def _setup_neutron_router(self):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            router = l3_models.Router()
            self.ctx.session.add(router)
            return router

    def _setup_distributed_binding(self, network_id,
                                   port_id, router_id, host_id):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            record = models.DistributedPortBinding(
                port_id=port_id,
                host=host_id,
                router_id=router_id,
                vif_type=portbindings.VIF_TYPE_UNBOUND,
                vnic_type=portbindings.VNIC_NORMAL,
                status='DOWN')
            self.ctx.session.add(record)
            return record

    def _setup_port_binding_level(self, network_id, port_id, host_id,
                                  level=0, driver='openvswitch'):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            record = models.PortBindingLevel(
                port_id=port_id,
                host=host_id,
                level=level,
                driver=driver,
                segment_id=network_id)
            self.ctx.session.add(record)
            return record

    def _setup_neutron_network_segment(self, segment_id, network_id):
        with db_api.CONTEXT_WRITER.using(self.ctx):
            segment = network_obj.NetworkSegment(
                self.ctx,
                id=segment_id, network_id=network_id,
                network_type=constants.TYPE_FLAT)
            segment.create()

    def test_ensure_distributed_port_binding_deals_with_db_duplicate(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        router_id = 'foo_router_id'
        host_id = 'foo_host_id'
        self._setup_neutron_network(network_id, [port_id])
        self._setup_distributed_binding(network_id, port_id,
                                        router_id, host_id)
        with mock.patch.object(query.Query, 'first') as query_first:
            query_first.return_value = []
            with mock.patch.object(ml2_db.LOG, 'debug') as log_trace:
                binding = ml2_db.ensure_distributed_port_binding(
                    self.ctx, port_id, host_id, router_id)
        self.assertTrue(query_first.called)
        self.assertTrue(log_trace.called)
        self.assertEqual(port_id, binding.port_id)

    def test_ensure_distributed_port_binding(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        self._setup_neutron_network(network_id, [port_id])
        router = self._setup_neutron_router()
        ml2_db.ensure_distributed_port_binding(
            self.ctx, port_id, 'foo_host', router.id)
        expected = (self.ctx.session.query(models.DistributedPortBinding).
                    filter_by(port_id=port_id).one())
        self.assertEqual(port_id, expected.port_id)

    def test_ensure_distributed_port_binding_multiple_bindings(self):
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        self._setup_neutron_network(network_id, [port_id])
        router = self._setup_neutron_router()
        ml2_db.ensure_distributed_port_binding(
            self.ctx, port_id, 'foo_host_1', router.id)
        ml2_db.ensure_distributed_port_binding(
            self.ctx, port_id, 'foo_host_2', router.id)
        bindings = (self.ctx.session.query(models.DistributedPortBinding).
                    filter_by(port_id=port_id).all())
        self.assertEqual(2, len(bindings))

    def test_delete_distributed_port_binding_if_stale(self):
        network_id = uuidutils.generate_uuid()
        segment_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        host_id = 'foo_host_id'
        self._setup_neutron_network(network_id, [port_id])
        self._setup_neutron_network_segment(segment_id, network_id)

        binding = self._setup_distributed_binding(
            network_id, port_id, None, host_id)
        binding_levels = self._setup_port_binding_level(
            segment_id, port_id, host_id)

        ml2_db.delete_distributed_port_binding_if_stale(
            self.ctx, binding)
        count = (self.ctx.session.query(models.DistributedPortBinding).
                 filter_by(port_id=binding.port_id, host=host_id).count())
        self.assertFalse(count)

        count = (self.ctx.session.query(models.PortBindingLevel).
                 filter_by(port_id=binding_levels.port_id,
                           host=host_id).count())
        self.assertFalse(count)

    def test_update_distributed_port_binding_by_host(self):
        network_id = uuidutils.generate_uuid()
        router_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        host_id = 'foo_host_id'
        self._setup_neutron_network(network_id, [port_id])

        binding = self._setup_distributed_binding(
            network_id, port_id, router_id, host_id)

        ml2_db.update_distributed_port_binding_by_host(
            self.ctx, binding.port_id, host_id, None)
        binding = (self.ctx.session.query(models.DistributedPortBinding).
                   filter_by(port_id=port_id, host=host_id).one())
        self.assertFalse(binding.router_id)

    def test_get_distributed_port_binding_by_host_not_found(self):
        port = ml2_db.get_distributed_port_binding_by_host(
            self.ctx, 'foo_port_id', 'foo_host_id')
        self.assertIsNone(port)

    def test_get_distributed_port_bindings_not_found(self):
        port = ml2_db.get_distributed_port_bindings(self.ctx,
                                                    'foo_port_id')
        self.assertFalse(len(port))

    def test_get_distributed_port_bindings(self):
        network_id = uuidutils.generate_uuid()
        port_id_1 = uuidutils.generate_uuid()
        port_id_2 = uuidutils.generate_uuid()
        self._setup_neutron_network(network_id, [port_id_1, port_id_2])
        router = self._setup_neutron_router()
        self._setup_distributed_binding(
            network_id, port_id_1, router.id, 'foo_host_id_1')
        self._setup_distributed_binding(
            network_id, port_id_1, router.id, 'foo_host_id_2')
        ports = ml2_db.get_distributed_port_bindings(self.ctx,
                                                     port_id_1)
        self.assertEqual(2, len(ports))

    def test_distributed_port_binding_deleted_by_port_deletion(self):
        network_id = uuidutils.generate_uuid()
        network_obj.Network(self.ctx, id=network_id).create()
        device_owner = constants.DEVICE_OWNER_DVR_INTERFACE
        port = port_obj.Port(
            self.ctx,
            id=uuidutils.generate_uuid(),
            network_id=network_id,
            mac_address=netaddr.EUI('00-11-22-33-44-55'),
            admin_state_up=True,
            status=constants.PORT_STATUS_ACTIVE,
            device_id='device_id',
            device_owner=device_owner)
        port.create()
        port_obj.DistributedPortBinding(
            self.ctx,
            port_id=port.id,
            host='host',
            vif_type=portbindings.VIF_TYPE_UNBOUND,
            vnic_type=portbindings.VNIC_NORMAL,
            router_id='router_id',
            status=constants.PORT_STATUS_DOWN).create()
        port_obj.DistributedPortBinding(
            self.ctx,
            port_id=port.id,
            host='another-host',
            vif_type=portbindings.VIF_TYPE_UNBOUND,
            vnic_type=portbindings.VNIC_NORMAL,
            router_id='router_id',
            status=constants.PORT_STATUS_DOWN).create()
        with warnings.catch_warnings(record=True) as warning_list:
            port.delete()
            self.assertEqual(
                [], warning_list,
                'Warnings: %s' % ';'.join([str(w) for w in warning_list]))
        ports = ml2_db.get_distributed_port_bindings(self.ctx,
                                                     port.id)
        self.assertEqual(0, len(ports))



# Copyright (c) 2021 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib import constants
from neutron_lib.utils import net as net_utils
from os_ken.lib import addrconv
from os_ken.lib.packet import dhcp
from os_ken.lib.packet import ethernet
from os_ken.lib.packet import in_proto as inet
from os_ken.lib.packet import ipv4
from os_ken.lib.packet import packet
from os_ken.lib.packet import udp

from neutron.agent.common import ovs_lib
from neutron.agent.l2.extensions.dhcp import base as dhcp_resp_base
from neutron.objects import subnet as subnet_obj
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_api
from neutron.tests import base


class FakeOF(object):
    OFPR_NO_MATCH = 0
    OFPR_ACTION = 1
    FPR_INVALID_TTL = 2
    OFP_NO_BUFFER = 3
    OFPP_CONTROLLER = 4


class FakeDatapath(object):
    ofproto = FakeOF()


class FakeMsg(object):
    datapath = FakeDatapath()
    reason = datapath.ofproto.OFPR_ACTION
    match = {'in_port': 1}
    data = ""
    buffer_id = 1
    total_len = 1
    table_id = 77
    cookie = 1

    def set_data(self, packet):
        packet.serialize()
        self.data = packet.data


PORT_INFO = {
    'device_owner': 'compute:nova',
    'admin_state_up': True,
    'network_id': 'd666ccb3-69e9-46cb-b157-bb3741d87d5a',
    'fixed_ips': [
        {'version': 4,
         'host_routes': [
             subnet_obj.Route(
                 destination=netaddr.IPNetwork('1.1.1.0/24'),
                 nexthop='192.168.1.100',
                 subnet_id='daed3c3d-d95a-48a8-a8b1-17d408cd760f'),
             subnet_obj.Route(
                 destination=netaddr.IPNetwork('2.2.2.2/32'),
                 nexthop='192.168.1.101',
                 subnet_id='daed3c3d-d95a-48a8-a8b1-17d408cd760f')],
         'subnet_id': 'daed3c3d-d95a-48a8-a8b1-17d408cd760f',
         'dns_nameservers': [
             subnet_obj.DNSNameServer(
                 address='8.8.8.8',
                 order=0,
                 subnet_id='daed3c3d-d95a-48a8-a8b1-17d408cd760f'),
             subnet_obj.DNSNameServer(
                 address='8.8.4.4',
                 order=1,
                 subnet_id='daed3c3d-d95a-48a8-a8b1-17d408cd760f')],
         'cidr': net_utils.AuthenticIPNetwork('192.168.111.0/24'),
         'ip_address': '192.168.111.45',
         'gateway_ip': netaddr.IPAddress('192.168.111.1')},
        {'version': 6,
         'host_routes': [],
         'subnet_id': 'bd013460-b05f-4927-a4c6-5127584b2487',
         'dns_nameservers': [],
         'cidr': net_utils.AuthenticIPNetwork('fda7:a5cc:3460:1::/64'),
         'ip_address': 'fda7:a5cc:3460:1::bf',
         'gateway_ip': netaddr.IPAddress('fda7:a5cc:3460:1::1')}
    ],
    'mac_address': '00:01:02:03:04:05',
    'port_id': '9a0e1889-f05f-43c7-a319-e1a723ed1587',
    'mtu': 1450
}


class DHCPResponderBaseTestCase(base.BaseTestCase):

    def setUp(self):
        super(DHCPResponderBaseTestCase, self).setUp()
        self.int_br = mock.Mock()
        self.tun_br = mock.Mock()
        self.plugin_rpc = mock.Mock()
        self.remote_resource_cache = mock.Mock()
        self.plugin_rpc.remote_resource_cache = self.remote_resource_cache
        self.agent_api = ovs_ext_api.OVSAgentExtensionAPI(
            self.int_br,
            self.tun_br,
            phys_brs=None,
            plugin_rpc=self.plugin_rpc)
        self.ext_api = mock.Mock()
        self.base_responer = dhcp_resp_base.DHCPResponderBase(self.agent_api,
                                                              self.ext_api)
        self.base_responer.int_br = self.int_br

        self.base_responer.get_dhcp_options = mock.Mock()
        self.base_responer.get_ret_packet = mock.Mock()
        self.base_responer.get_state = mock.Mock()
        self.base_responer.handle_dhcp = mock.Mock()

        self.port_info = PORT_INFO

    def _create_test_dhcp_request_packet(self):
        option_list = []
        bin_server = addrconv.ipv4.text_to_bin('192.168.1.1')
        option_list.append(
            dhcp.option(tag=dhcp.DHCP_SERVER_IDENTIFIER_OPT,
                        value=bin_server))
        option_list.append(
                dhcp.option(tag=dhcp.DHCP_MESSAGE_TYPE_OPT,
                            value=b'\x03'))
        options = dhcp.options(option_list=option_list)
        ret_pkt = packet.Packet()
        ret_pkt.add_protocol(ethernet.ethernet(
            dst="ff:ff:ff:ff:ff:ff",
            src=self.port_info['mac_address']))
        ret_pkt.add_protocol(
            ipv4.ipv4(dst="255.255.255.255",
                      src="0.0.0.0",
                      proto=inet.IPPROTO_UDP))
        ret_pkt.add_protocol(udp.udp(
            src_port=constants.DHCP_CLIENT_PORT,
            dst_port=constants.DHCP_RESPONSE_PORT))
        ret_pkt.add_protocol(dhcp.dhcp(op=dhcp.DHCP_BOOT_REQUEST,
                                       chaddr=self.port_info['mac_address'],
                                       siaddr='0.0.0.0',
                                       xid=3454038351,
                                       options=options))
        return ret_pkt

    def test__packet_in_handler(self):
        ev = mock.Mock()
        ev.msg = FakeMsg()
        dhcp_req = self._create_test_dhcp_request_packet()
        ev.msg.set_data(dhcp_req)

        self.base_responer.get_port_id_from_br = mock.Mock()
        self.base_responer.get_port_id_from_br.return_value = (
            PORT_INFO['port_id'])
        self.ext_api.get_port_info = mock.Mock()
        self.ext_api.get_port_info.return_value = PORT_INFO
        self.base_responer._packet_in_handler(ev)
        self.base_responer.handle_dhcp.assert_called_once_with(
            ev.msg.datapath, 1, mock.ANY, self.port_info)

    def test_get_bin_dns(self):
        except_value = b'\x08\x08\x08\x08\x08\x08\x04\x04'
        bin_dns = self.base_responer.get_bin_dns(
            self.port_info['fixed_ips'][0]['dns_nameservers'])
        self.assertEqual(except_value, bin_dns)

    def test_get_bin_route(self):
        expect_bin_route = b' \x08\x08\x08\x08\xc0\xa8\x01\x01'
        bin_route = self.base_responer.get_bin_route('8.8.8.8', '192.168.1.1')
        self.assertEqual(expect_bin_route, bin_route)

    def test_get_port_id_from_br(self):
        self.int_br.get_vif_ports = mock.Mock()
        self.int_br.get_vif_ports.return_value = [
            ovs_lib.VifPort(port_name="tap-1", ofport=1,
                            vif_id=PORT_INFO['port_id'],
                            vif_mac=PORT_INFO['mac_address'],
                            switch='br-int')]
        pid = self.base_responer.get_port_id_from_br(
            1, PORT_INFO['mac_address'])
        self.assertEqual(PORT_INFO['port_id'], pid)

    def test_get_port_ip(self):
        ip_v4 = self.base_responer.get_port_ip(self.port_info, 4)
        self.assertEqual(self.port_info['fixed_ips'][0],
                         ip_v4)
        ip_v6 = self.base_responer.get_port_ip(self.port_info, 6)
        self.assertEqual(self.port_info['fixed_ips'][1],
                         ip_v6)

    def test_packet_out(self):
        datapath = mock.Mock()
        ofport = 1
        dhcp_req = self._create_test_dhcp_request_packet()
        self.base_responer.packet_out(datapath, ofport, dhcp_req)
        datapath.send_msg.assert_called_once_with(mock.ANY)



# Copyright (c) 2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
from unittest import mock
import urllib

from neutron_lib.api import attributes
from neutron_lib.api import converters
from neutron_lib.api.definitions import empty_string_filtering
from neutron_lib.api.definitions import filter_validation
from neutron_lib.callbacks import registry
from neutron_lib import constants
from neutron_lib import context
from neutron_lib import exceptions as n_exc
from neutron_lib import fixture
from neutron_lib.tests.unit import fake_notifier
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_policy import policy as oslo_policy
from oslo_utils import uuidutils
import webob
from webob import exc
import webtest

from neutron.api import api_common
from neutron.api import extensions
from neutron.api.v2 import base as v2_base
from neutron.api.v2 import router
from neutron.conf import quota as quota_conf
from neutron import policy
from neutron import quota
from neutron.tests import base
from neutron.tests import tools
from neutron.tests.unit import testlib_api


EXTDIR = os.path.join(base.ROOTDIR, 'unit/extensions')
NULL_QUOTA_DRIVER = 'neutron.db.quota.driver_null.DbQuotaDriverNull'

_uuid = uuidutils.generate_uuid


def _get_path(resource, id=None, action=None,
              fmt=None, endpoint=None):
    path = '/%s' % resource

    if id is not None:
        path = path + '/%s' % id

    if action is not None:
        path = path + '/%s' % action

    if endpoint is not None:
        path = path + '/%s' % endpoint

    if fmt is not None:
        path = path + '.%s' % fmt

    return path


def _get_neutron_env(tenant_id=None, as_admin=False):
    tenant_id = tenant_id or _uuid()
    roles = ['member', 'reader']
    if as_admin:
        roles.append('admin')
    return {'neutron.context': context.Context('', tenant_id, roles=roles)}


class APIv2TestBase(base.BaseTestCase):
    def setUp(self):
        super(APIv2TestBase, self).setUp()

        plugin = 'neutron.neutron_plugin_base_v2.NeutronPluginBaseV2'
        # Ensure existing ExtensionManager is not used
        extensions.PluginAwareExtensionManager._instance = None
        # Create the default configurations
        self.config_parse()
        # Update the plugin
        self.setup_coreplugin(plugin, load_plugins=False)
        self._plugin_patcher = mock.patch(plugin, autospec=True)
        self.plugin = self._plugin_patcher.start()
        instance = self.plugin.return_value
        instance.supported_extension_aliases = [empty_string_filtering.ALIAS,
                                                filter_validation.ALIAS]
        instance._NeutronPluginBaseV2__native_pagination_support = True
        instance._NeutronPluginBaseV2__native_sorting_support = True
        instance._NeutronPluginBaseV2__filter_validation_support = True
        tools.make_mock_plugin_json_encodable(instance)

        api = router.APIRouter()
        self.api = webtest.TestApp(api)

        self._tenant_id = "api-test-tenant"

        quota.QUOTAS._driver = None
        cfg.CONF.set_override('quota_driver', quota_conf.QUOTA_DB_DRIVER,
                              group='QUOTAS')

        # APIRouter initialization resets policy module, re-initializing it
        policy.init()

    def _post_request(self, path, initial_input, expect_errors=None,
                      req_tenant_id=None, as_admin=False):
        req_tenant_id = req_tenant_id or self._tenant_id
        return self.api.post_json(
            path, initial_input, expect_errors=expect_errors,
            extra_environ=_get_neutron_env(req_tenant_id, as_admin))

    def _put_request(self, path, initial_input, expect_errors=None,
                     req_tenant_id=None, as_admin=False):
        req_tenant_id = req_tenant_id or self._tenant_id
        return self.api.put_json(
            path, initial_input, expect_errors=expect_errors,
            extra_environ=_get_neutron_env(req_tenant_id, as_admin))

    def _delete_request(self, path, expect_errors=None,
                        req_tenant_id=None, as_admin=False):
        req_tenant_id = req_tenant_id or self._tenant_id
        return self.api.delete_json(
            path, expect_errors=expect_errors,
            extra_environ=_get_neutron_env(req_tenant_id, as_admin))


class _ArgMatcher(object):
    """An adapter to assist mock assertions, used to custom compare."""

    def __init__(self, cmp, obj):
        self.cmp = cmp
        self.obj = obj

    def __eq__(self, other):
        return self.cmp(self.obj, other)


def _list_cmp(l1, l2):
    return set(l1) == set(l2)


class APIv2TestCase(APIv2TestBase):

    @staticmethod
    def _get_policy_attrs(attr_info):
        policy_attrs = {name for (name, info) in attr_info.items()
                        if info.get('required_by_policy')}
        if 'tenant_id' in policy_attrs:
            policy_attrs.add('project_id')
        return sorted(policy_attrs)

    def _do_field_list(self, resource, base_fields):
        attr_info = attributes.RESOURCES[resource]
        policy_attrs = self._get_policy_attrs(attr_info)
        for name, info in attr_info.items():
            if info.get('primary_key'):
                policy_attrs.append(name)
        fields = base_fields
        fields.extend(policy_attrs)
        return fields

    def _get_collection_kwargs(self, skipargs=None, **kwargs):
        skipargs = skipargs or []
        args_list = ['filters', 'fields', 'sorts', 'limit', 'marker',
                     'page_reverse']
        args_dict = dict(
            (arg, mock.ANY) for arg in set(args_list) - set(skipargs))
        args_dict.update(kwargs)
        return args_dict

    def test_fields(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'fields': 'foo'})
        fields = self._do_field_list('networks', ['foo'])
        kwargs = self._get_collection_kwargs(fields=fields)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_fields_multiple(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        fields = self._do_field_list('networks', ['bar', 'foo'])
        self.api.get(_get_path('networks'), {'fields': ['foo', 'bar']})
        kwargs = self._get_collection_kwargs(fields=fields)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_fields_multiple_with_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        fields = self._do_field_list('networks', ['foo'])
        self.api.get(_get_path('networks'), {'fields': ['foo', '']})
        kwargs = self._get_collection_kwargs(fields=fields)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_fields_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'fields': ''})
        kwargs = self._get_collection_kwargs(fields=[])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_fields_multiple_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'fields': ['', '']})
        kwargs = self._get_collection_kwargs(fields=[])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': 'bar'})
        filters = {'name': ['bar']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': ''})
        filters = {'name': ['']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_multiple_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': ['', '']})
        filters = {'name': ['', '']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_multiple_with_empty(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': ['bar', '']})
        filters = {'name': ['bar', '']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_multiple_values(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': ['bar', 'bar2']})
        filters = {'name': ['bar', 'bar2']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_multiple(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': 'bar',
                                             'tenant_id': 'bar2'})
        filters = {'name': ['bar'], 'tenant_id': ['bar2']}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_with_fields(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'name': 'bar', 'fields': 'foo'})
        filters = {'name': ['bar']}
        fields = self._do_field_list('networks', ['foo'])
        kwargs = self._get_collection_kwargs(filters=filters, fields=fields)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_with_convert_to(self):
        instance = self.plugin.return_value
        instance.get_ports.return_value = []

        self.api.get(_get_path('ports'), {'admin_state_up': 'true'})
        filters = {'admin_state_up': [True]}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_ports.assert_called_once_with(mock.ANY, **kwargs)

    def test_filters_with_convert_list_to(self):
        instance = self.plugin.return_value
        instance.get_ports.return_value = []

        self.api.get(_get_path('ports'),
                     {'fixed_ips': ['ip_address=foo', 'subnet_id=bar']})
        filters = {'fixed_ips': {'ip_address': ['foo'], 'subnet_id': ['bar']}}
        kwargs = self._get_collection_kwargs(filters=filters)
        instance.get_ports.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'limit': '10'})
        kwargs = self._get_collection_kwargs(limit=10)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_great_than_max_limit(self):
        cfg.CONF.set_default('pagination_max_limit', '1000')
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'limit': '1001'})
        kwargs = self._get_collection_kwargs(limit=1000)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_zero(self):
        cfg.CONF.set_default('pagination_max_limit', '1000')
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'), {'limit': '0'})
        kwargs = self._get_collection_kwargs(limit=1000)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_unspecific(self):
        cfg.CONF.set_default('pagination_max_limit', '1000')
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'))
        kwargs = self._get_collection_kwargs(limit=1000)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_negative_value(self):
        cfg.CONF.set_default('pagination_max_limit', '1000')
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        res = self.api.get(_get_path('networks'), {'limit': -1},
                           expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_limit_with_non_integer(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        res = self.api.get(_get_path('networks'),
                           {'limit': 'abc'}, expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)
        self.assertIn('abc', res)

    def test_limit_with_infinite_pagination_max_limit(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []
        cfg.CONF.set_override('pagination_max_limit', 'Infinite')
        self.api.get(_get_path('networks'))
        kwargs = self._get_collection_kwargs(limit=None)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_negative_pagination_max_limit(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []
        cfg.CONF.set_default('pagination_max_limit', '-1')
        self.api.get(_get_path('networks'))
        kwargs = self._get_collection_kwargs(limit=None)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_limit_with_non_integer_pagination_max_limit(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []
        cfg.CONF.set_default('pagination_max_limit', 'abc')
        self.api.get(_get_path('networks'))
        kwargs = self._get_collection_kwargs(limit=None)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_marker(self):
        cfg.CONF.set_override('pagination_max_limit', '1000')
        instance = self.plugin.return_value
        instance.get_networks.return_value = []
        marker = _uuid()
        self.api.get(_get_path('networks'),
                     {'marker': marker})
        kwargs = self._get_collection_kwargs(limit=1000, marker=marker)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_page_reverse(self):
        calls = []
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'page_reverse': 'True'})
        kwargs = self._get_collection_kwargs(page_reverse=True)
        calls.append(mock.call.get_networks(mock.ANY, **kwargs))
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

        instance.get_networks.reset_mock()

        self.api.get(_get_path('networks'),
                     {'page_reverse': 'False'})
        kwargs = self._get_collection_kwargs(page_reverse=False)
        calls.append(mock.call.get_networks(mock.ANY, **kwargs))
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_page_reverse_with_non_bool(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'page_reverse': 'abc'})
        kwargs = self._get_collection_kwargs(page_reverse=False)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_page_reverse_with_unspecific(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'))
        kwargs = self._get_collection_kwargs(page_reverse=False)
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_sort(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'sort_key': ['name', 'admin_state_up'],
                      'sort_dir': ['desc', 'asc']})
        kwargs = self._get_collection_kwargs(sorts=[('name', False),
                                                    ('admin_state_up', True),
                                                    ('id', True)])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_sort_with_primary_key(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        self.api.get(_get_path('networks'),
                     {'sort_key': ['name', 'admin_state_up', 'id'],
                      'sort_dir': ['desc', 'asc', 'desc']})
        kwargs = self._get_collection_kwargs(sorts=[('name', False),
                                                    ('admin_state_up', True),
                                                    ('id', False)])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_sort_without_direction(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        res = self.api.get(_get_path('networks'), {'sort_key': ['name']},
                           expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_sort_with_invalid_attribute(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        res = self.api.get(_get_path('networks'),
                           {'sort_key': 'abc',
                            'sort_dir': 'asc'},
                           expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_sort_with_invalid_dirs(self):
        instance = self.plugin.return_value
        instance.get_networks.return_value = []

        res = self.api.get(_get_path('networks'),
                           {'sort_key': 'name',
                            'sort_dir': 'abc'},
                           expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_emulated_sort(self):
        instance = self.plugin.return_value
        instance._NeutronPluginBaseV2__native_pagination_support = False
        instance._NeutronPluginBaseV2__native_sorting_support = False
        instance.get_networks.return_value = []
        api = webtest.TestApp(router.APIRouter())
        api.get(_get_path('networks'), {'sort_key': ['name', 'status'],
                                        'sort_dir': ['desc', 'asc']})
        kwargs = self._get_collection_kwargs(
            skipargs=['sorts', 'limit', 'marker', 'page_reverse'])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_emulated_sort_without_sort_field(self):
        instance = self.plugin.return_value
        instance._NeutronPluginBaseV2__native_pagination_support = False
        instance._NeutronPluginBaseV2__native_sorting_support = False
        instance.get_networks.return_value = []
        api = webtest.TestApp(router.APIRouter())
        api.get(_get_path('networks'), {'sort_key': ['name', 'status'],
                                        'sort_dir': ['desc', 'asc'],
                                        'fields': ['subnets']})
        kwargs = self._get_collection_kwargs(
            skipargs=['sorts', 'limit', 'marker', 'page_reverse'],
            fields=_ArgMatcher(_list_cmp, ['name',
                                           'status',
                                           'id',
                                           'subnets',
                                           'shared',
                                           'project_id',
                                           'tenant_id']))
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_emulated_pagination(self):
        instance = self.plugin.return_value
        instance._NeutronPluginBaseV2__native_pagination_support = False
        instance.get_networks.return_value = []
        api = webtest.TestApp(router.APIRouter())
        api.get(_get_path('networks'), {'limit': 10,
                                        'marker': 'foo',
                                        'page_reverse': False})
        kwargs = self._get_collection_kwargs(skipargs=['limit',
                                                       'marker',
                                                       'page_reverse'])
        instance.get_networks.assert_called_once_with(mock.ANY, **kwargs)

    def test_native_pagination_without_native_sorting(self):
        instance = self.plugin.return_value
        instance._NeutronPluginBaseV2__native_sorting_support = False
        self.assertRaises(n_exc.Invalid, router.APIRouter)


# Note: since all resources use the same controller and validation
# logic, we actually get really good coverage from testing just networks.
class JSONV2TestCase(APIv2TestBase, testlib_api.WebTestCase):

    def _test_list(self, req_tenant_id, real_tenant_id):
        env = {}
        if req_tenant_id:
            env = _get_neutron_env(req_tenant_id)
        input_dict = {'id': uuidutils.generate_uuid(),
                      'name': 'net1',
                      'admin_state_up': True,
                      'status': "ACTIVE",
                      'project_id': real_tenant_id,
                      'shared': False,
                      'subnets': []}
        instance = self.plugin.return_value
        instance.get_networks.return_value = [input_dict]

        res = self.api.get(_get_path('networks',
                                     fmt=self.fmt), extra_environ=env)
        res = self.deserialize(res)
        self.assertIn('networks', res)
        if not req_tenant_id or req_tenant_id == real_tenant_id:
            # expect full list returned
            self.assertEqual(1, len(res['networks']))
            output_dict = res['networks'][0]
            input_dict['shared'] = False
            self.assertEqual(len(input_dict), len(output_dict))
            for k, v in input_dict.items():
                self.assertEqual(v, output_dict[k])
        else:
            # expect no results
            self.assertEqual(0, len(res['networks']))

    def test_list_noauth(self):
        self._test_list(None, _uuid())

    def test_list_keystone(self):
        tenant_id = _uuid()
        self._test_list(tenant_id, tenant_id)

    def test_list_keystone_bad(self):
        tenant_id = _uuid()
        self._test_list(tenant_id + "bad", tenant_id)

    def test_list_pagination(self):
        id1 = str(_uuid())
        id2 = str(_uuid())
        input_dict1 = {'id': id1,
                       'name': 'net1',
                       'admin_state_up': True,
                       'status': "ACTIVE",
                       'tenant_id': '',
                       'shared': False,
                       'subnets': []}
        input_dict2 = {'id': id2,
                       'name': 'net2',
                       'admin_state_up': True,
                       'status': "ACTIVE",
                       'tenant_id': '',
                       'shared': False,
                       'subnets': []}
        return_value = [input_dict1, input_dict2]
        instance = self.plugin.return_value
        instance.get_networks.return_value = return_value
        params = {'limit': ['2'],
                  'marker': [str(_uuid())],
                  'sort_key': ['name'],
                  'sort_dir': ['asc']}
        res = self.api.get(_get_path('networks'),
                           params=params).json

        self.assertEqual(2, len(res['networks']))
        self.assertEqual(sorted([id1, id2]),
                         sorted([res['networks'][0]['id'],
                                res['networks'][1]['id']]))

        self.assertIn('networks_links', res)
        next_links = []
        previous_links = []
        for r in res['networks_links']:
            if r['rel'] == 'next':
                next_links.append(r)
            if r['rel'] == 'previous':
                previous_links.append(r)
        self.assertEqual(1, len(next_links))
        self.assertEqual(1, len(previous_links))

        url = urllib.parse.urlparse(next_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        params['marker'] = [id2]
        self.assertEqual(params, urllib.parse.parse_qs(url.query))

        url = urllib.parse.urlparse(previous_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        params['marker'] = [id1]
        params['page_reverse'] = ['True']
        self.assertEqual(params, urllib.parse.parse_qs(url.query))

    def test_list_pagination_with_last_page(self):
        id = str(_uuid())
        input_dict = {'id': id,
                      'name': 'net1',
                      'admin_state_up': True,
                      'status': "ACTIVE",
                      'tenant_id': '',
                      'shared': False,
                      'subnets': []}
        return_value = [input_dict]
        instance = self.plugin.return_value
        instance.get_networks.return_value = return_value
        params = {'limit': ['2'],
                  'marker': str(_uuid())}
        res = self.api.get(_get_path('networks'),
                           params=params).json

        self.assertEqual(1, len(res['networks']))
        self.assertEqual(id, res['networks'][0]['id'])

        self.assertIn('networks_links', res)
        previous_links = []
        for r in res['networks_links']:
            self.assertNotEqual(r['rel'], 'next')
            if r['rel'] == 'previous':
                previous_links.append(r)
        self.assertEqual(1, len(previous_links))

        url = urllib.parse.urlparse(previous_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        expect_params = params.copy()
        expect_params['marker'] = [id]
        expect_params['page_reverse'] = ['True']
        self.assertEqual(expect_params, urllib.parse.parse_qs(url.query))

    def test_list_pagination_with_empty_page(self):
        return_value = []
        instance = self.plugin.return_value
        instance.get_networks.return_value = return_value
        params = {'limit': ['2'],
                  'marker': str(_uuid())}
        res = self.api.get(_get_path('networks'),
                           params=params).json

        self.assertEqual([], res['networks'])

        previous_links = []
        if 'networks_links' in res:
            for r in res['networks_links']:
                self.assertNotEqual(r['rel'], 'next')
                if r['rel'] == 'previous':
                    previous_links.append(r)
        self.assertEqual(1, len(previous_links))

        url = urllib.parse.urlparse(previous_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        expect_params = params.copy()
        del expect_params['marker']
        expect_params['page_reverse'] = ['True']
        self.assertEqual(expect_params, urllib.parse.parse_qs(url.query))

    def test_list_pagination_reverse_with_last_page(self):
        id = str(_uuid())
        input_dict = {'id': id,
                      'name': 'net1',
                      'admin_state_up': True,
                      'status': "ACTIVE",
                      'tenant_id': '',
                      'shared': False,
                      'subnets': []}
        return_value = [input_dict]
        instance = self.plugin.return_value
        instance.get_networks.return_value = return_value
        params = {'limit': ['2'],
                  'marker': [str(_uuid())],
                  'page_reverse': ['True']}
        res = self.api.get(_get_path('networks'),
                           params=params).json

        self.assertEqual(len(res['networks']), 1)
        self.assertEqual(id, res['networks'][0]['id'])

        self.assertIn('networks_links', res)
        next_links = []
        for r in res['networks_links']:
            self.assertNotEqual(r['rel'], 'previous')
            if r['rel'] == 'next':
                next_links.append(r)
        self.assertEqual(1, len(next_links))

        url = urllib.parse.urlparse(next_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        expected_params = params.copy()
        del expected_params['page_reverse']
        expected_params['marker'] = [id]
        self.assertEqual(expected_params,
                         urllib.parse.parse_qs(url.query))

    def test_list_pagination_reverse_with_empty_page(self):
        return_value = []
        instance = self.plugin.return_value
        instance.get_networks.return_value = return_value
        params = {'limit': ['2'],
                  'marker': [str(_uuid())],
                  'page_reverse': ['True']}
        res = self.api.get(_get_path('networks'),
                           params=params).json
        self.assertEqual([], res['networks'])

        next_links = []
        if 'networks_links' in res:
            for r in res['networks_links']:
                self.assertNotEqual(r['rel'], 'previous')
                if r['rel'] == 'next':
                    next_links.append(r)
        self.assertEqual(1, len(next_links))

        url = urllib.parse.urlparse(next_links[0]['href'])
        self.assertEqual(url.path, _get_path('networks'))
        expect_params = params.copy()
        del expect_params['marker']
        del expect_params['page_reverse']
        self.assertEqual(expect_params, urllib.parse.parse_qs(url.query))

    def test_create(self):
        net_id = _uuid()
        data = {'network': {'name': 'net1', 'admin_state_up': True,
                            'tenant_id': _uuid()}}
        return_value = {'subnets': [], 'status': "ACTIVE",
                        'id': net_id}
        return_value.update(data['network'].copy())

        instance = self.plugin.return_value
        instance.create_network.return_value = return_value
        instance.get_networks_count.return_value = 0

        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/' + self.fmt)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('network', res)
        net = res['network']
        self.assertEqual(net_id, net['id'])
        self.assertEqual("ACTIVE", net['status'])

    def test_create_use_defaults(self):
        net_id = _uuid()
        tenant_id = _uuid()

        initial_input = {'network': {'name': 'net1',
                                     'tenant_id': tenant_id,
                                     'project_id': tenant_id}}
        full_input = {'network': {'admin_state_up': True,
                                  'shared': False}}
        full_input['network'].update(initial_input['network'])

        return_value = {'id': net_id, 'status': "ACTIVE"}
        return_value.update(full_input['network'])

        instance = self.plugin.return_value
        instance.create_network.return_value = return_value
        instance.get_networks_count.return_value = 0

        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(initial_input),
                            content_type='application/' + self.fmt)
        instance.create_network.assert_called_with(mock.ANY,
                                                   network=full_input)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('network', res)
        net = res['network']
        self.assertEqual(net_id, net['id'])
        self.assertTrue(net['admin_state_up'])
        self.assertEqual("ACTIVE", net['status'])

    def test_create_no_keystone_env(self):
        data = {'name': 'net1'}
        self._test_create_failure_bad_request('networks', data)

    def test_create_with_keystone_env(self):
        tenant_id = _uuid()
        net_id = _uuid()
        env = _get_neutron_env(tenant_id)
        # tenant_id should be fetched from env
        initial_input = {'network': {'name': 'net1'}}
        full_input = {'network': {'admin_state_up': True,
                      'shared': False, 'tenant_id': tenant_id,
                      'project_id': tenant_id}}
        full_input['network'].update(initial_input['network'])

        return_value = {'id': net_id, 'status': "ACTIVE"}
        return_value.update(full_input['network'])

        instance = self.plugin.return_value
        instance.create_network.return_value = return_value
        instance.get_networks_count.return_value = 0

        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(initial_input),
                            content_type='application/' + self.fmt,
                            extra_environ=env)

        instance.create_network.assert_called_with(mock.ANY,
                                                   network=full_input)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)

    def test_create_bad_keystone_tenant(self):
        tenant_id = _uuid()
        data = {'network': {'name': 'net1', 'tenant_id': tenant_id}}
        env = {'neutron.context': context.Context('', tenant_id + "bad")}
        self._test_create_failure_bad_request('networks', data,
                                              extra_environ=env)

    def test_create_no_body(self):
        data = {'whoa': None}
        self._test_create_failure_bad_request('networks', data)

    def test_create_body_string_not_json(self):
        data = 'a string'
        self._test_create_failure_bad_request('networks', data)

    def test_create_body_boolean_not_json(self):
        data = True
        self._test_create_failure_bad_request('networks', data)

    def test_create_no_resource(self):
        data = {}
        self._test_create_failure_bad_request('networks', data)

    def test_create_object_string_not_json(self):
        data = {'network': 'a string'}
        self._test_create_failure_bad_request('networks', data)

    def test_create_object_boolean_not_json(self):
        data = {'network': True}
        self._test_create_failure_bad_request('networks', data)

    def test_create_missing_attr(self):
        data = {'port': {'what': 'who', 'tenant_id': _uuid()}}
        self._test_create_failure_bad_request('ports', data)

    def test_create_readonly_attr(self):
        data = {'network': {'name': 'net1', 'tenant_id': _uuid(),
                            'status': "ACTIVE"}}
        self._test_create_failure_bad_request('networks', data)

    def test_create_with_too_long_name(self):
        data = {'network': {'name': "12345678" * 32,
                            'admin_state_up': True,
                            'tenant_id': _uuid()}}
        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/' + self.fmt,
                            expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_create_bulk(self):
        data = {'networks': [{'name': 'net1',
                              'admin_state_up': True,
                              'tenant_id': _uuid()},
                             {'name': 'net2',
                              'admin_state_up': True,
                              'tenant_id': _uuid()}]}

        def side_effect(context, network):
            net = network.copy()
            net['network'].update({'subnets': []})
            return net['network']

        instance = self.plugin.return_value
        instance.create_network.side_effect = side_effect
        instance.get_networks_count.return_value = 0
        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/' + self.fmt)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)

    def _test_create_failure_bad_request(self, resource, data, **kwargs):
        res = self.api.post(_get_path(resource, fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/' + self.fmt,
                            expect_errors=True, **kwargs)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_create_bulk_networks_none(self):
        self._test_create_failure_bad_request('networks', {'networks': None})

    def test_create_bulk_networks_empty_list(self):
        self._test_create_failure_bad_request('networks', {'networks': []})

    def test_create_bulk_missing_attr(self):
        data = {'ports': [{'what': 'who', 'tenant_id': _uuid()}]}
        self._test_create_failure_bad_request('ports', data)

    def test_create_bulk_partial_body(self):
        data = {'ports': [{'device_id': 'device_1',
                           'tenant_id': _uuid()},
                          {'tenant_id': _uuid()}]}
        self._test_create_failure_bad_request('ports', data)

    def test_create_attr_not_specified(self):
        net_id = _uuid()
        tenant_id = _uuid()
        device_id = _uuid()
        initial_input = {'port': {'name': '', 'network_id': net_id,
                                  'tenant_id': tenant_id,
                                  'project_id': tenant_id,
                                  'device_id': device_id,
                                  'admin_state_up': True}}
        full_input = {'port': {'admin_state_up': True,
                               'mac_address': constants.ATTR_NOT_SPECIFIED,
                               'fixed_ips': constants.ATTR_NOT_SPECIFIED,
                               'device_owner': ''}}
        full_input['port'].update(initial_input['port'])
        return_value = {'id': _uuid(), 'status': 'ACTIVE',
                        'admin_state_up': True,
                        'mac_address': 'ca:fe:de:ad:be:ef',
                        'device_id': device_id,
                        'device_owner': ''}
        return_value.update(initial_input['port'])

        instance = self.plugin.return_value
        instance.get_network.return_value = {
            'tenant_id': str(tenant_id)
        }
        instance.get_ports_count.return_value = 1
        instance.create_port.return_value = return_value
        res = self.api.post(_get_path('ports', fmt=self.fmt),
                            self.serialize(initial_input),
                            content_type='application/' + self.fmt)
        instance.create_port.assert_called_with(mock.ANY, port=full_input)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('port', res)
        port = res['port']
        self.assertEqual(net_id, port['network_id'])
        self.assertEqual('ca:fe:de:ad:be:ef', port['mac_address'])

    def test_create_return_extra_attr(self):
        net_id = _uuid()
        project_id = _uuid()
        data = {'network': {'name': 'net1', 'admin_state_up': True,
                            'tenant_id': project_id}}
        return_value = {'subnets': [], 'status': "ACTIVE",
                        'id': net_id, 'v2attrs:something': "123"}
        return_value.update(data['network'].copy())

        instance = self.plugin.return_value
        instance.create_network.return_value = return_value
        instance.get_networks_count.return_value = 0

        res = self.api.post(_get_path('networks', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/' + self.fmt,
                            extra_environ=_get_neutron_env(project_id))
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('network', res)
        net = res['network']
        self.assertEqual(net_id, net['id'])
        self.assertEqual("ACTIVE", net['status'])
        self.assertNotIn('v2attrs:something', net)

    def test_fields(self):
        project_id = _uuid()
        return_value = {'name': 'net1', 'admin_state_up': True,
                        'project_id': project_id, 'subnets': []}

        instance = self.plugin.return_value
        instance.get_network.return_value = return_value

        self.api.get(_get_path('networks',
                               id=uuidutils.generate_uuid(),
                               fmt=self.fmt),
                     extra_environ=_get_neutron_env(project_id))

    def _test_delete(self, req_tenant_id, real_tenant_id, expected_code,
                     expect_errors=False):
        env = {}
        if req_tenant_id:
            env = _get_neutron_env(req_tenant_id)
        instance = self.plugin.return_value
        instance.get_network.return_value = {'project_id': real_tenant_id,
                                             'shared': False}
        instance.delete_network.return_value = None

        res = self.api.delete(_get_path('networks',
                                        id=uuidutils.generate_uuid(),
                                        fmt=self.fmt),
                              extra_environ=env,
                              expect_errors=expect_errors)
        self.assertEqual(expected_code, res.status_int)

    def test_delete_noauth(self):
        self._test_delete(None, _uuid(), exc.HTTPNoContent.code)

    def test_delete_keystone(self):
        tenant_id = _uuid()
        self._test_delete(tenant_id, tenant_id, exc.HTTPNoContent.code)

    def test_delete_keystone_bad_tenant(self):
        tenant_id = _uuid()
        self._test_delete(tenant_id + "bad", tenant_id,
                          exc.HTTPNotFound.code, expect_errors=True)

    def _test_get(self, req_tenant_id, real_tenant_id, expected_code,
                  expect_errors=False):
        shared = req_tenant_id and req_tenant_id.endswith('another')
        env = {}
        if req_tenant_id:
            env = _get_neutron_env(req_tenant_id)

        data = {'project_id': real_tenant_id, 'shared': shared}
        instance = self.plugin.return_value
        instance.get_network.return_value = data

        res = self.api.get(_get_path('networks',
                                     id=uuidutils.generate_uuid(),
                                     fmt=self.fmt),
                           extra_environ=env,
                           expect_errors=expect_errors)
        self.assertEqual(expected_code, res.status_int)
        return res

    def test_get_noauth(self):
        self._test_get(None, _uuid(), 200)

    def test_get_keystone(self):
        tenant_id = _uuid()
        self._test_get(tenant_id, tenant_id, 200)

    def test_get_keystone_bad_tenant(self):
        tenant_id = _uuid()
        self._test_get(tenant_id + "bad", tenant_id,
                       exc.HTTPNotFound.code, expect_errors=True)

    def test_get_keystone_shared_network(self):
        tenant_id = _uuid()
        self._test_get(tenant_id + "another", tenant_id, 200)

    def test_get_keystone_strip_admin_only_attribute(self):
        tenant_id = _uuid()
        # Inject rule in policy engine
        rules = oslo_policy.Rules.from_dict(
            {'get_network:name': "rule:admin_only"})
        policy.set_rules(rules, overwrite=False)
        res = self._test_get(tenant_id, tenant_id, 200)
        res = self.deserialize(res)
        self.assertNotIn('name', res['network'])

    def _test_update(self, req_tenant_id, real_tenant_id, expected_code,
                     expect_errors=False):
        env = {}
        if req_tenant_id:
            env = _get_neutron_env(req_tenant_id)
        # leave out 'name' field intentionally
        data = {'network': {'admin_state_up': True}}
        return_value = {'subnets': []}
        return_value.update(data['network'].copy())

        instance = self.plugin.return_value
        instance.get_network.return_value = {'project_id': real_tenant_id,
                                             'shared': False}
        instance.update_network.return_value = return_value

        res = self.api.put(_get_path('networks',
                                     id=uuidutils.generate_uuid(),
                                     fmt=self.fmt),
                           self.serialize(data),
                           extra_environ=env,
                           expect_errors=expect_errors)
        #  Ensure id attribute is included in fields returned by GET call
        #  in update procedure.
        self.assertEqual(1, instance.get_network.call_count)
        self.assertIn('id', instance.get_network.call_args[1]['fields'])
        self.assertEqual(res.status_int, expected_code)

    def test_update_noauth(self):
        self._test_update(None, _uuid(), 200)

    def test_update_keystone(self):
        tenant_id = _uuid()
        self._test_update(tenant_id, tenant_id, 200)

    def test_update_keystone_bad_tenant(self):
        tenant_id = _uuid()
        self._test_update(tenant_id + "bad", tenant_id,
                          exc.HTTPNotFound.code, expect_errors=True)

    def test_update_keystone_no_tenant(self):
        tenant_id = _uuid()
        self._test_update(tenant_id, None,
                          exc.HTTPNotFound.code, expect_errors=True)

    def test_update_readonly_field(self):
        data = {'network': {'status': "NANANA"}}
        res = self.api.put(_get_path('networks', id=_uuid()),
                           self.serialize(data),
                           content_type='application/' + self.fmt,
                           expect_errors=True)
        self.assertEqual(400, res.status_int)

    def test_invalid_attribute_field(self):
        data = {'network': {'invalid_key1': "foo1", 'invalid_key2': "foo2"}}
        res = self.api.put(_get_path('networks', id=_uuid()),
                           self.serialize(data),
                           content_type='application/' + self.fmt,
                           expect_errors=True)
        self.assertEqual(400, res.status_int)

    def test_retry_on_index(self):
        instance = self.plugin.return_value
        instance.get_networks.side_effect = [db_exc.RetryRequest(None), []]
        api = webtest.TestApp(router.APIRouter())
        api.get(_get_path('networks', fmt=self.fmt))
        self.assertTrue(instance.get_networks.called)

    def test_retry_on_show(self):
        instance = self.plugin.return_value
        instance.get_network.side_effect = [db_exc.RetryRequest(None), {}]
        api = webtest.TestApp(router.APIRouter())
        api.get(_get_path('networks', _uuid(), fmt=self.fmt))
        self.assertTrue(instance.get_network.called)


# Note: since all resources use the same controller and validation
# logic, we actually get really good coverage from testing just networks.
class V2Views(base.BaseTestCase):
    def _view(self, keys, collection, resource):
        data = dict((key, 'value') for key in keys)
        data['fake'] = 'value'
        attr_info = attributes.RESOURCES[collection]
        controller = v2_base.Controller(None, collection, resource, attr_info)
        res = controller._view(context.get_admin_context(), data)
        self.assertNotIn('fake', res)
        for key in keys:
            self.assertIn(key, res)

    def test_network(self):
        keys = ('id', 'name', 'subnets', 'admin_state_up', 'status',
                'tenant_id')
        self._view(keys, 'networks', 'network')

    def test_port(self):
        keys = ('id', 'network_id', 'mac_address', 'fixed_ips',
                'device_id', 'admin_state_up', 'tenant_id', 'status')
        self._view(keys, 'ports', 'port')

    def test_subnet(self):
        keys = ('id', 'network_id', 'tenant_id', 'gateway_ip',
                'ip_version', 'cidr', 'enable_dhcp')
        self._view(keys, 'subnets', 'subnet')


class NotificationTest(APIv2TestBase):

    def setUp(self):
        super(NotificationTest, self).setUp()
        fake_notifier.reset()
        quota.QUOTAS._driver = None
        cfg.CONF.set_override('quota_driver', NULL_QUOTA_DRIVER,
                              group='QUOTAS')

    def _resource_op_notifier(self, opname, resource, expected_errors=False):
        tenant_id = _uuid()
        network_obj = {'name': 'myname',
                       'project_id': tenant_id}
        initial_input = {resource: network_obj}
        instance = self.plugin.return_value
        instance.get_network.return_value = network_obj
        instance.get_networks_count.return_value = 0
        expected_code = exc.HTTPCreated.code
        if opname == 'create':
            res = self._post_request(
                _get_path('networks'),
                initial_input, expect_errors=expected_errors,
                req_tenant_id=tenant_id)
        if opname == 'update':
            op_input = {resource: {'name': 'myname'}}
            res = self._put_request(
                _get_path('networks', id=tenant_id),
                op_input, expect_errors=expected_errors,
                req_tenant_id=tenant_id)
            expected_code = exc.HTTPOk.code
        if opname == 'delete':
            res = self._delete_request(
                _get_path('networks', id=tenant_id),
                expect_errors=expected_errors,
                req_tenant_id=tenant_id)
            expected_code = exc.HTTPNoContent.code

        expected_events = ('.'.join([resource, opname, "start"]),
                           '.'.join([resource, opname, "end"]))
        self.assertEqual(len(expected_events),
                         len(fake_notifier.NOTIFICATIONS))
        for msg, event in zip(fake_notifier.NOTIFICATIONS, expected_events):
            self.assertEqual('INFO', msg['priority'])
            self.assertEqual(event, msg['event_type'])
            if opname == 'delete' and event == 'network.delete.end':
                self.assertIn('payload', msg)
                resource = msg['payload']
                self.assertIn('network_id', resource)
                self.assertIn('network', resource)

        self.assertEqual(expected_code, res.status_int)

    def test_network_create_notifier(self):
        self._resource_op_notifier('create', 'network')

    def test_network_delete_notifier(self):
        self._resource_op_notifier('delete', 'network')

    def test_network_update_notifier(self):
        self._resource_op_notifier('update', 'network')


class RegistryNotificationTest(APIv2TestBase):

    def setUp(self):
        super(RegistryNotificationTest, self).setUp()
        quota.QUOTAS._driver = None
        cfg.CONF.set_override('quota_driver', NULL_QUOTA_DRIVER,
                              group='QUOTAS')

    def _test_registry_publish(self, opname, resource, initial_input=None):
        instance = self.plugin.return_value
        instance.get_networks.return_value = initial_input
        instance.get_networks_count.return_value = 0
        expected_code = exc.HTTPCreated.code
        with mock.patch.object(registry, 'publish') as publish:
            if opname == 'create':
                res = self.api.post_json(
                    _get_path('networks'),
                    initial_input)
            if opname == 'update':
                res = self.api.put_json(
                    _get_path('networks', id=_uuid()),
                    initial_input)
                expected_code = exc.HTTPOk.code
            if opname == 'delete':
                res = self.api.delete(_get_path('networks', id=_uuid()))
                expected_code = exc.HTTPNoContent.code
            self.assertTrue(publish.called)
        self.assertEqual(expected_code, res.status_int)

    def test_network_create_registry_publish(self):
        input = {'network': {'name': 'net',
                             'tenant_id': _uuid()}}
        self._test_registry_publish('create', 'network', input)

    def test_network_delete_registry_publish(self):
        self._test_registry_publish('delete', 'network')

    def test_network_update_registry_publish(self):
        input = {'network': {'name': 'net'}}
        self._test_registry_publish('update', 'network', input)

    def test_networks_create_bulk_registry_publish(self):
        input = {'networks': [{'name': 'net1',
                               'tenant_id': _uuid()},
                              {'name': 'net2',
                               'tenant_id': _uuid()}]}
        self._test_registry_publish('create', 'network', input)


class QuotaTest(APIv2TestBase):
    """This class checks the quota enforcement API, regardless of the driver"""

    def test_create_network_quota_exceeded(self):
        initial_input = {'network': {'name': 'net1', 'tenant_id': _uuid()}}
        with mock.patch.object(quota.QUOTAS, 'make_reservation',
                               side_effect=n_exc.OverQuota(overs='network')):
            res = self.api.post_json(
                _get_path('networks'), initial_input, expect_errors=True)
        self.assertIn("Quota exceeded for resources",
                      res.json['NeutronError']['message'])

    def test_create_network_quota_without_limit(self):
        initial_input = {'network': {'name': 'net1', 'tenant_id': _uuid()}}
        with mock.patch.object(quota.QUOTAS, 'make_reservation'), \
                mock.patch.object(quota.QUOTAS, 'commit_reservation'):
            res = self.api.post_json(
                _get_path('networks'), initial_input)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)


class ExtensionTestCase(base.BaseTestCase):
    def setUp(self):
        super(ExtensionTestCase, self).setUp()
        plugin = 'neutron.neutron_plugin_base_v2.NeutronPluginBaseV2'
        # Ensure existing ExtensionManager is not used
        extensions.PluginAwareExtensionManager._instance = None

        self.useFixture(fixture.APIDefinitionFixture())

        # Create the default configurations
        self.config_parse()

        # Update the plugin and extensions path
        self.setup_coreplugin(plugin, load_plugins=False)
        cfg.CONF.set_override('api_extensions_path', EXTDIR)

        self._plugin_patcher = mock.patch(plugin, autospec=True)
        self.plugin = self._plugin_patcher.start()

        # Instantiate mock plugin and enable the V2attributes extension
        self.plugin.return_value.supported_extension_aliases = ["v2attrs"]

        api = router.APIRouter()
        self.api = webtest.TestApp(api)

        quota.QUOTAS._driver = None
        cfg.CONF.set_override('quota_driver', NULL_QUOTA_DRIVER,
                              group='QUOTAS')

    def test_extended_create(self):
        net_id = _uuid()
        tenant_id = _uuid()
        initial_input = {'network': {'name': 'net1', 'tenant_id': tenant_id,
                                     'project_id': tenant_id,
                                     'v2attrs:something_else': "abc"}}
        data = {'network': {'admin_state_up': True, 'shared': False}}
        data['network'].update(initial_input['network'])

        return_value = {'subnets': [], 'status': "ACTIVE",
                        'id': net_id,
                        'v2attrs:something': "123"}
        return_value.update(data['network'].copy())

        instance = self.plugin.return_value
        instance.create_network.return_value = return_value
        instance.get_networks_count.return_value = 0

        res = self.api.post_json(
            _get_path('networks'), initial_input,
            extra_environ=_get_neutron_env(tenant_id))

        instance.create_network.assert_called_with(mock.ANY,
                                                   network=data)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        self.assertIn('network', res.json)
        net = res.json['network']
        self.assertEqual(net_id, net['id'])
        self.assertEqual("ACTIVE", net['status'])
        self.assertEqual("123", net['v2attrs:something'])
        self.assertNotIn('v2attrs:something_else', net)


class TestSubresourcePlugin(object):
    def get_network_dummies(self, context, network_id,
                            filters=None, fields=None):
        return []

    def get_network_dummy(self, context, id, network_id,
                          fields=None):
        return {}

    def create_network_dummy(self, context, network_id, dummy):
        return {}

    def update_network_dummy(self, context, id, network_id, dummy):
        return {}

    def delete_network_dummy(self, context, id, network_id):
        return

    def mactions(self, context, id, network_id):
        return


class ListArgsTestCase(base.BaseTestCase):
    def test_list_args(self):
        path = '/?fields=4&foo=3&fields=2&bar=1'
        request = webob.Request.blank(path)
        expect_val = ['2', '4']
        actual_val = api_common.list_args(request, 'fields')
        self.assertEqual(expect_val, sorted(actual_val))

    def test_list_args_with_empty(self):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        self.assertEqual([], api_common.list_args(request, 'fields'))


class SortingTestCase(base.BaseTestCase):
    def test_get_sorts(self):
        path = '/?sort_key=foo&sort_dir=desc&sort_key=bar&sort_dir=asc'
        request = webob.Request.blank(path)
        attr_info = {
            'foo': {'key': 'val', 'is_sort_key': True},
            'bar': {'key': 'val', 'is_sort_key': True}
        }
        expect_val = [('foo', False), ('bar', True)]
        actual_val = api_common.get_sorts(request, attr_info)
        self.assertEqual(expect_val, actual_val)

    def test_get_sorts_with_project_id(self):
        path = '/?sort_key=project_id&sort_dir=desc'
        request = webob.Request.blank(path)
        attr_info = {'tenant_id': {'key': 'val', 'is_sort_key': True}}
        expect_val = [('project_id', False)]
        actual_val = api_common.get_sorts(request, attr_info)
        self.assertEqual(expect_val, actual_val)

    def test_get_sorts_with_non_sort_key(self):
        path = '/?sort_key=created_at&sort_dir=desc'
        request = webob.Request.blank(path)
        attr_info = {
            'foo': {'key': 'val', 'is_sort_key': True},
            'bar': {'key': 'val', 'is_sort_key': True},
            'created_at': {'key': 'val'}
        }
        self.assertRaises(exc.HTTPBadRequest,
                          api_common.get_sorts,
                          request, attr_info)


class FiltersTestCase(base.BaseTestCase):
    def test_all_skip_args(self):
        path = '/?fields=4&fields=3&fields=2&fields=1'
        request = webob.Request.blank(path)
        self.assertEqual({}, api_common.get_filters(request, {},
                                                    ["fields"]))

    @mock.patch('neutron.api.api_common.is_empty_string_filtering_supported',
                return_value=False)
    def test_blank_values(self, mock_is_supported):
        path = '/?foo=&bar=&baz=&qux='
        request = webob.Request.blank(path)
        self.assertEqual({}, api_common.get_filters(request, {}))

    @mock.patch('neutron.api.api_common.is_empty_string_filtering_supported',
                return_value=True)
    def test_blank_values_with_filtering_supported(self, mock_is_supported):
        path = '/?foo=&bar=&baz=&qux='
        request = webob.Request.blank(path)
        self.assertEqual({'foo': [''], 'bar': [''], 'baz': [''], 'qux': ['']},
                         api_common.get_filters(request, {}))

    def test_no_attr_info(self):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        actual_val = api_common.get_filters(request, {})
        self.assertEqual(expect_val, actual_val)

    def test_attr_info_with_project_info_populated(self):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        attr_info = {'tenant_id': {'key': 'val'}}
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        actual_val = api_common.get_filters(request, attr_info)
        self.assertEqual(expect_val, actual_val)
        expect_attr_info = {'tenant_id': {'key': 'val'},
                            'project_id': {'key': 'val'}}
        self.assertEqual(expect_attr_info, attr_info)

    @mock.patch('neutron.api.api_common.is_filter_validation_enabled',
                return_value=True)
    def test_attr_info_with_filter_validation(self, mock_validation_enabled):
        attr_info = {}
        self._test_attr_info(attr_info)

        attr_info = {'foo': {}}
        self._test_attr_info(attr_info)

        attr_info = {'foo': {'is_filter': False}}
        self._test_attr_info(attr_info)

        attr_info = {'foo': {'is_filter': False}, 'bar': {'is_filter': True},
                     'baz': {'is_filter': True}, 'qux': {'is_filter': True}}
        self._test_attr_info(attr_info)

        attr_info = {'foo': {'is_filter': True}, 'bar': {'is_filter': True},
                     'baz': {'is_filter': True}, 'qux': {'is_filter': True}}
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        self._test_attr_info(attr_info, expect_val)

        attr_info = {'foo': {'is_filter': True}, 'bar': {'is_filter': True},
                     'baz': {'is_filter': True}, 'qux': {'is_filter': True},
                     'quz': {}}
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        self._test_attr_info(attr_info, expect_val)

        attr_info = {'foo': {'is_filter': True}, 'bar': {'is_filter': True},
                     'baz': {'is_filter': True}, 'qux': {'is_filter': True},
                     'quz': {'is_filter': False}}
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        self._test_attr_info(attr_info, expect_val)

    def _test_attr_info(self, attr_info, expect_val=None):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        if expect_val:
            actual_val = api_common.get_filters(
                request, attr_info,
                is_filter_validation_supported=True)
            self.assertEqual(expect_val, actual_val)
        else:
            self.assertRaises(
                exc.HTTPBadRequest, api_common.get_filters, request, attr_info,
                is_filter_validation_supported=True)

    def test_attr_info_without_conversion(self):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        attr_info = {'foo': {'key': 'val'}}
        expect_val = {'foo': ['4'], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        actual_val = api_common.get_filters(request, attr_info)
        self.assertEqual(expect_val, actual_val)

    def test_attr_info_with_convert_list_to(self):
        path = '/?foo=key=4&bar=3&foo=key=2&qux=1'
        request = webob.Request.blank(path)
        attr_info = {
            'foo': {
                'convert_list_to': converters.convert_kvp_list_to_dict,
            }
        }
        expect_val = {'foo': {'key': ['2', '4']}, 'bar': ['3'], 'qux': ['1']}
        actual_val = api_common.get_filters(request, attr_info)
        self.assertOrderedEqual(expect_val, actual_val)

    def test_attr_info_with_convert_to(self):
        path = '/?foo=4&bar=3&baz=2&qux=1'
        request = webob.Request.blank(path)
        attr_info = {'foo': {'convert_to': converters.convert_to_int}}
        expect_val = {'foo': [4], 'bar': ['3'], 'baz': ['2'], 'qux': ['1']}
        actual_val = api_common.get_filters(request, attr_info)
        self.assertEqual(expect_val, actual_val)

    def test_attr_info_with_base_db_attributes(self):
        path = '/?__contains__=1&__class__=2'
        request = webob.Request.blank(path)
        self.assertEqual({}, api_common.get_filters(request, {}))


class CreateResourceTestCase(base.BaseTestCase):
    def test_resource_creation(self):
        resource = v2_base.create_resource('fakes', 'fake', None, {})
        self.assertIsInstance(resource, webob.dec.wsgify)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import constants as db_const
from neutron_lib.db import model_base
import sqlalchemy as sa
from sqlalchemy import sql

from neutron.db import rbac_db_models


class AddressScope(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a neutron address scope."""

    __tablename__ = "address_scopes"

    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE), nullable=False)

    # TODO(imalinovskiy): drop this field when contract migrations will be
    # allowed again
    # NOTE(imalinovskiy): this field cannot be removed from model due to
    # functional test test_models_sync, trailing underscore is required to
    # prevent conflicts with RBAC code
    shared_ = sa.Column("shared", sa.Boolean, nullable=False,
                        server_default=sql.false())

    ip_version = sa.Column(sa.Integer(), nullable=False)

    rbac_entries = sa.orm.relationship(rbac_db_models.AddressScopeRBAC,
                                       backref='address_scopes',
                                       lazy='joined',
                                       cascade='all, delete, delete-orphan')



# Copyright (c) 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron_lib import context
from neutron_lib import exceptions

from neutron.db import db_base_plugin_v2 as base_plugin
from neutron.db.quota import api as quota_api
from neutron.db.quota import driver
from neutron.objects import quota as quota_obj
from neutron.quota import resource
from neutron.tests import base
from neutron.tests.unit import quota as test_quota
from neutron.tests.unit import testlib_api

DB_PLUGIN_KLASS = 'neutron.db.db_base_plugin_v2.NeutronDbPluginV2'


def _count_resource(context, resource, project_id):
    """A fake counting function to determine current used counts"""
    if resource[-1] == 's':
        resource = resource[:-1]
    result = quota_obj.QuotaUsage.get_object_dirty_protected(
        context, resource=resource)
    return 0 if not result else result.in_use


class FakePlugin(base_plugin.NeutronDbPluginV2, driver.DbQuotaDriver):
    """A fake plugin class containing all DB methods."""


class TestResource(object):
    """Describe a test resource for quota checking."""

    def __init__(self, name, default, fake_count=0):
        self.name = name
        self.quota = default
        self.fake_count = fake_count

    @property
    def default(self):
        return self.quota

    def count(self, *args, **kwargs):
        return self.fake_count


class TestTrackedResource(resource.TrackedResource):
    """Describes a test tracked resource for detailed quota checking"""
    def __init__(self, name, model_class, flag=None,
                 plural_name=None):
        super(TestTrackedResource, self).__init__(
            name, model_class, flag=flag, plural_name=None)

    @property
    def default(self):
        return self.flag


class TestCountableResource(resource.CountableResource):
    """Describes a test countable resource for detailed quota checking"""
    def __init__(self, name, count, flag=-1, plural_name=None):
        super(TestCountableResource, self).__init__(
            name, count, flag=flag, plural_name=None)

    @property
    def default(self):
        return self.flag


PROJECT = 'prj_test'
RESOURCE = 'res_test'
ALT_RESOURCE = 'res_test_meh'


class TestDbQuotaDriver(testlib_api.SqlTestCase,
                        base.BaseTestCase):
    def setUp(self):
        super(TestDbQuotaDriver, self).setUp()
        self.plugin = FakePlugin()
        self.context = context.get_admin_context()
        self.setup_coreplugin(core_plugin=DB_PLUGIN_KLASS)
        self.quota_driver = driver.DbQuotaDriver()
        self.project_1, self.project_2 = 'prj_test_1', 'prj_test_2'
        self.resource_1, self.resource_2 = 'res_test_1', 'res_test_2'
        self.projects = (self.project_1, self.project_2)
        self.resources = (self.resource_1, self.resource_2)

    def test_create_quota_limit(self):
        defaults = {RESOURCE: TestResource(RESOURCE, 4)}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        quotas = self.plugin.get_project_quotas(self.context, defaults,
                                                PROJECT)
        self.assertEqual(2, quotas[RESOURCE])

    def test_update_quota_limit(self):
        defaults = {RESOURCE: TestResource(RESOURCE, 4)}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 3)
        quotas = self.plugin.get_project_quotas(self.context, defaults,
                                                PROJECT)
        self.assertEqual(3, quotas[RESOURCE])

    def test_delete_project_quota_restores_default_limit(self):
        defaults = {RESOURCE: TestResource(RESOURCE, 4)}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.plugin.delete_project_quota(self.context, PROJECT)
        quotas = self.plugin.get_project_quotas(self.context, defaults,
                                                PROJECT)
        self.assertEqual(4, quotas[RESOURCE])

    def test_get_default_quotas(self):
        defaults = {RESOURCE: TestResource(RESOURCE, 4)}
        user_ctx = context.Context(user_id=PROJECT, tenant_id=PROJECT)
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        quotas = self.plugin.get_default_quotas(user_ctx, defaults, PROJECT)
        self.assertEqual(4, quotas[RESOURCE])

    def test_get_project_quotas(self):
        user_ctx = context.Context(user_id=PROJECT, tenant_id=PROJECT)
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        quotas = self.plugin.get_project_quotas(user_ctx, {}, PROJECT)
        self.assertEqual(2, quotas[RESOURCE])

    def test_get_project_quotas_different_project(self):
        user_ctx = context.Context(user_id=PROJECT,
                                   tenant_id='another_project')
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        # It is appropriate to use assertFalse here as the expected return
        # value is an empty dict (the defaults passed in the statement below
        # after the request context)
        self.assertFalse(self.plugin.get_project_quotas(user_ctx, {}, PROJECT))

    def test_get_all_quotas(self):
        resources = {self.resource_1: TestResource(self.resource_1, 3),
                     self.resource_2: TestResource(self.resource_2, 5)}

        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_1, 7)
        self.plugin.update_quota_limit(self.context, self.project_2,
                                       self.resource_2, 9)
        quotas = self.plugin.get_all_quotas(self.context, resources)

        # Expect two projects' quotas
        self.assertEqual(2, len(quotas))
        # But not quotas for the same project twice
        self.assertNotEqual(quotas[0]['project_id'], quotas[1]['project_id'])

        # Check the expected limits. The quotas can be in any order.
        for quota in quotas:
            project = quota['project_id']
            self.assertIn(project, (self.project_1, self.project_2))
            if project == self.project_1:
                expected_limit_r1 = 7
                expected_limit_r2 = 5
            if project == self.project_2:
                expected_limit_r1 = 3
                expected_limit_r2 = 9
            self.assertEqual(expected_limit_r1, quota[self.resource_1])
            self.assertEqual(expected_limit_r2, quota[self.resource_2])

    def test_limit_check(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2)}
        values = {RESOURCE: 1}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.plugin.limit_check(self.context, PROJECT, resources, values)

    def test_limit_check_over_quota(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2)}
        values = {RESOURCE: 3}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)

        self.assertRaises(exceptions.OverQuota, self.plugin.limit_check,
                          context.get_admin_context(), PROJECT, resources,
                          values)

    def test_limit_check_equals_to_quota(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2)}
        values = {RESOURCE: 2}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.plugin.limit_check(self.context, PROJECT, resources, values)

    def test_limit_check_value_lower_than_zero(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2)}
        values = {RESOURCE: -1}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.assertRaises(exceptions.InvalidQuotaValue,
                          self.plugin.limit_check, context.get_admin_context(),
                          PROJECT, resources, values)

    def _test_make_reservation_success(self, quota_driver,
                                       resource_name, deltas):
        resources = {resource_name: TestResource(resource_name, 2)}
        self.plugin.update_quota_limit(self.context, PROJECT, resource_name, 2)
        reservation = quota_driver.make_reservation(
            self.context,
            self.context.project_id,
            resources,
            deltas,
            self.plugin)
        self.assertIn(resource_name, reservation.deltas)
        self.assertEqual(deltas[resource_name],
                         reservation.deltas[resource_name])
        self.assertEqual(self.context.project_id,
                         reservation.project_id)

    def test_make_reservation_single_resource(self):
        self._test_make_reservation_success(
            self.quota_driver, RESOURCE, {RESOURCE: 1})

    def test_make_reservation_fill_quota(self):
        self._test_make_reservation_success(
            self.quota_driver, RESOURCE, {RESOURCE: 2})

    def test_make_reservation_multiple_resources(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2),
                     ALT_RESOURCE: TestResource(ALT_RESOURCE, 2)}
        deltas = {RESOURCE: 1, ALT_RESOURCE: 2}
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.plugin.update_quota_limit(self.context, PROJECT, ALT_RESOURCE, 2)
        reservation = self.quota_driver.make_reservation(
            self.context,
            self.context.project_id,
            resources,
            deltas,
            self.plugin)
        self.assertIn(RESOURCE, reservation.deltas)
        self.assertIn(ALT_RESOURCE, reservation.deltas)
        self.assertEqual(1, reservation.deltas[RESOURCE])
        self.assertEqual(2, reservation.deltas[ALT_RESOURCE])
        self.assertEqual(self.context.project_id,
                         reservation.project_id)

    def test_make_reservation_over_quota_fails(self):
        resources = {RESOURCE: TestResource(RESOURCE, 2,
                                            fake_count=2)}
        deltas = {RESOURCE: 1}
        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 2)
        self.assertRaises(exceptions.OverQuota,
                          self.quota_driver.make_reservation,
                          self.context,
                          self.context.project_id,
                          resources,
                          deltas,
                          self.plugin)

    def test_get_detailed_project_quotas_resource(self):
        res = {RESOURCE: TestTrackedResource(RESOURCE, test_quota.MehModel)}

        self.plugin.update_quota_limit(self.context, PROJECT, RESOURCE, 6)
        self.quota_driver.make_reservation(self.context, PROJECT, res,
                                           {RESOURCE: 1}, self.plugin)
        quota_api.set_quota_usage(self.context, RESOURCE, PROJECT, 2)
        detailed_quota = self.plugin.get_detailed_project_quotas(
            self.context, res, PROJECT)
        self.assertEqual(6, detailed_quota[RESOURCE]['limit'])
        self.assertEqual(2, detailed_quota[RESOURCE]['used'])
        self.assertEqual(1, detailed_quota[RESOURCE]['reserved'])

    def _create_resources(self):
        return {
            self.resource_1:
                TestTrackedResource(self.resource_1, test_quota.MehModel),
            self.resource_2:
                TestCountableResource(self.resource_2, _count_resource)}

    def test_get_detailed_project_quotas_multiple_resource(self):
        resources = self._create_resources()
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_1, 6)
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_2, 9)
        self.quota_driver.make_reservation(
            self.context, self.project_1, resources,
            {self.resource_1: 1, self.resource_2: 7}, self.plugin)

        quota_api.set_quota_usage(self.context, self.resource_1,
                                  self.project_1, 2)
        quota_api.set_quota_usage(self.context, self.resource_2,
                                  self.project_1, 3)
        detailed_quota = self.plugin.get_detailed_project_quotas(
            self.context, resources, self.project_1)

        self.assertEqual(6, detailed_quota[self.resource_1]['limit'])
        self.assertEqual(1, detailed_quota[self.resource_1]['reserved'])
        self.assertEqual(2, detailed_quota[self.resource_1]['used'])

        self.assertEqual(9, detailed_quota[self.resource_2]['limit'])
        self.assertEqual(7, detailed_quota[self.resource_2]['reserved'])
        self.assertEqual(3, detailed_quota[self.resource_2]['used'])

    def test_quota_limit_check(self):
        resources = self._create_resources()
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_1, 10)
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_2, 10)
        reservations = {self.resource_1: 8}
        self.quota_driver.make_reservation(
            self.context, self.project_1, resources, reservations, self.plugin)
        resources[self.resource_2]._count_func = lambda x, y, z: 8

        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources, {self.resource_1: 2}))
        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources, {self.resource_2: 2}))
        self.assertRaises(
            exceptions.OverQuota, self.quota_driver.quota_limit_check,
            self.context, self.project_1, resources, {self.resource_1: 3})
        self.assertRaises(
            exceptions.OverQuota, self.quota_driver.quota_limit_check,
            self.context, self.project_1, resources, {self.resource_2: 3})
        self.assertRaises(
            exceptions.OverQuota, self.quota_driver.quota_limit_check,
            self.context, self.project_1, resources,
            {self.resource_1: 3, self.resource_2: 3})

    def test_quota_limit_check_unlimited(self):
        resources = self._create_resources()
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_1, -1)
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_2, -1)
        reservations = {self.resource_1: 8}
        self.quota_driver.make_reservation(
            self.context, self.project_1, resources, reservations, self.plugin)
        resources[self.resource_2]._count_func = lambda x, y, z: 8

        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources, {self.resource_1: 2}))
        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources, {self.resource_2: 2}))
        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources,
            {self.resource_1: 10 ** 9, self.resource_2: 10 ** 9}))

    def test_quota_limit_check_untracked_resource(self):
        resources = self._create_resources()
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_1, -1)
        self.plugin.update_quota_limit(self.context, self.project_1,
                                       self.resource_2, -1)
        reservations = {self.resource_1: 8}
        self.quota_driver.make_reservation(
            self.context, self.project_1, resources, reservations, self.plugin)
        resources[self.resource_2]._count_func = lambda x, y, z: 8

        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources,
            {self.resource_1: 10 ** 9}))
        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources,
            {self.resource_2: 10 ** 9}))
        self.assertIsNone(self.quota_driver.quota_limit_check(
            self.context, self.project_1, resources, {'untracked': 10 ** 9}))
